/****************************************************************************/
/* ZEPTOOS:zepto-info */
/*     This file is part of ZeptoOS: The Small Linux for Big Computers.
 *     See www.mcs.anl.gov/zeptoos for more information.
 */
/* ZEPTOOS:zepto-info */
/* */
/* ZEPTOOS:zepto-fillin */
/*     $Id: buffers.c,v 1.9 2007/06/27 15:10:45 iskra Exp $
 *     ZeptoOS_Version: 1.2
 *     ZeptoOS_Heredity: FOSS_ORIG
 *     ZeptoOS_License: GPL
 */
/* ZEPTOOS:zepto-fillin */
/* */
/* ZEPTOOS:zepto-gpl */
/*      Copyright: Argonne National Laboratory, Department of Energy,
 *                 and UChicago Argonne, LLC.  2004, 2005, 2006, 2007
 *      ZeptoOS License: GPL
 * 
 *      This software is free.  See the file ZeptoOS/misc/license.GPL
 *      for complete details on your rights to copy, modify, and use this
 *      software.
 */
/* ZEPTOOS:zepto-gpl */
/****************************************************************************/

#define _GNU_SOURCE /* For posix_memalign.  */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

#include <errno.h>
#include <pthread.h>

#include "bgl.h"
#include "zoid.h"
#include "zoid_api.h"

static struct zoid_buffer* empty_queue_1 = NULL;
static struct zoid_buffer* empty_queue_2 = NULL;
static pthread_mutex_t empty_queue_mutex = PTHREAD_MUTEX_INITIALIZER;

void*
__zoid_allocate_buffer(size_t size)
{
    struct zoid_buffer* buffer;
    struct zoid_buffer** pqueue;

    if (size > max_buffer_size_2)
    {
	/* This simply indicates an error on the caller's side, and could
	   in principle take place only for the result buffer.  We return
	   a hard error, which is propagated to the caller, who can then
	   adjust the request size and resubmit.  */
	struct thread_specific_data* thread_data =
	    (struct thread_specific_data*)
	    pthread_getspecific(thread_specific_key);
	thread_data->errnum = E2BIG;
	thread_data->last_excessive_size = size - max_buffer_size_2;
	return NULL;
    }

    if (pthread_mutex_lock(&empty_queue_mutex))
    {
	perror("locking empty queue mutex");
	return NULL;
    }

    pqueue = (size > max_buffer_size_1 ?
	      &empty_queue_2 : &empty_queue_1);
    if (!*pqueue)
    {
	void* empty_queue_ptr = pqueue;

	if (posix_memalign(empty_queue_ptr, 16,
			   sizeof(struct zoid_buffer) +
			   (size > max_buffer_size_1 ? max_buffer_size_2 :
			    max_buffer_size_1)))
	{
	    /* Yuck!  There's nothing we can do...  */
	    struct thread_specific_data* thread_data =
		(struct thread_specific_data*)
		pthread_getspecific(thread_specific_key);
	    fprintf(stderr, "Failed to allocate memory for a buffer!\n");
	    thread_data->errnum = ENOMEM;
	    pthread_mutex_unlock(&empty_queue_mutex);
	    return NULL;
	}

	(*pqueue)->size1 = (size <= max_buffer_size_1);
	(*pqueue)->next = NULL;
    }

    buffer = *pqueue;
    *pqueue = (*pqueue)->next;

    if (pthread_mutex_unlock(&empty_queue_mutex))
    {
	perror("unlocking empty queue mutex");
	return NULL;
    }

    buffer->errnum = buffer->excessive_size = 0;
    buffer->userbuf_out = NULL;
    buffer->userbuf_out_cb = NULL;
    buffer->userbuf_out_priv = NULL;
    buffer->userbuf_in = NULL;
    buffer->ack_sent = 0;

    return buffer->data;
}

void
__zoid_release_buffer(void* buffer)
{
    struct zoid_buffer* entry;

    if (!buffer)
	return;

    if (pthread_mutex_lock(&empty_queue_mutex))
    {
	perror("locking empty queue mutex");
	return;
    }

    entry = get_zoid_buffer(buffer);

    if (entry->userbuf_out)
    {
	/* Note: we cannot call set_zoid_calling_process_id() here, because
	   entry->softheader might be uninitialized at this point.  */
	entry->userbuf_out_cb(entry->userbuf_out, entry->userbuf_out_priv);
	entry->userbuf_out = NULL;
    }

    if (entry->size1)
    {
	entry->next = empty_queue_1;
	empty_queue_1 = entry;
    }
    else
    {
	entry->next = empty_queue_2;
	empty_queue_2 = entry;
    }

    if (pthread_mutex_unlock(&empty_queue_mutex))
    {
	perror("unlocking empty queue mutex");
	return;
    }

    /* We might want to do something more sophisticated in the future, like
       release of buffers if the queue length exceeds some value, etc.  */
}

struct zoid_buffer*
get_zoid_buffer(void* buffer)
{
    if (buffer)
	buffer -= offsetof(struct zoid_buffer, data);
    return buffer;
}
