#!/usr/bin/perl -w
#***************************************************************************
# ZEPTOOS:zepto-info
#      This file is part of ZeptoOS: The Small Linux for Big Computers.
#      See www.mcs.anl.gov/zeptoos for more information.
# ZEPTOOS:zepto-info
#
# ZEPTOOS:zepto-fillin
#      $Id: scanner.pl,v 1.16 2007/06/27 15:10:45 iskra Exp $
#      ZeptoOS_Version: 1.2
#      ZeptoOS_Heredity: FOSS_ORIG
#      ZeptoOS_License: GPL
# ZEPTOOS:zepto-fillin
#
# ZEPTOOS:zepto-gpl
#       Copyright: Argonne National Laboratory, Department of Energy,
#                  and UChicago Argonne, LLC.  2004, 2005, 2006, 2007
#       ZeptoOS License: GPL
#  
#       This software is free.  See the file ZeptoOS/misc/license.GPL
#       for complete details on your rights to copy, modify, and use this
#       software.
# ZEPTOOS:zepto-gpl
#***************************************************************************
#
# Parses a C header file.  For every function prototype found, generates two
# output files: one for a client calling the function and one for the server,
# where the function is in fact executed.  Code in the generated files
# performs marshalling and demarshalling of function arguments and results.
#
# The parser is very limited, and does not handle many C constructs.  It is
# thus essential that the input header file be as simple as possible.
# In particular, function prototypes should be specified at the end of the
# file, not intermixed with any other specifications such as data type
# definitions.
#
# Ordinary comments are best placed on separate lines.
#
# The parser ignores everything until it encounters a magic line:
#     /* START-ZOID-SCANNER ID=<n> INIT=<s1> FINI=<s2> */
# <n> should be a small (16-bit) number, unique for every header file
# <s1> is the name of an init function from the implementation object that
#     will be called before function forwarding starts.  A single integer
#     argument will be passed to that function: the number of processes
#     that will be handled by that daemon instance.  If an init function
#     is not required, please specify INIT=NULL.
# <s2> is the name of a termination function from the implementation object
#     that will be called after all processes have exited.  No arguments
#     will be passed to that function.  If a termination function is not
#     required, please specify FINI=NULL.
#
# (please note that the parser is case *sensitive*).
#
# Hints are needed to correctly encode and decode function arguments.
# These are placed after each argument, before a separating comma (or a
# closing bracket), and are embedded inside dedicated C comments.  Multiple
# hints per argument are usually provided; these are separated by a colon.
# The following hints are currently defined:
# - in, out, inout: whether the argument is an input argument, an output
#   argument, or both.  "in" is the default.
# - obj, str, ptr, arr, arr2d: argument type -- either an object (say, an
#   "int", or a structure passed by value), a '\0'-terminated character
#   string, a pointer to an object, an array, or a two-dimensional array
#   (type**, not type[][]).  "obj" is the default.
# - size: used for array types -- indicates the index of the argument that
#   specifies the array size.  Absolute numbers are accepted (1 to <number of
#   arguments>) or relative ones ("+1" for next argument, "-1" for previous
#   argument, etc).  For one-dimensional array, the argument must be of a
#   numerical type, or a pointer to such type.  For two-dimensional arrays,
#   the argument must be a one-dimensional array of numerical elements,
#   specifying the sizes along the less significant dimension of the array
#   (the size of the more significant dimension is the size of the
#   one-dimensional array itself).  Please note that the unit of size is the
#   size of the base array type (thus, size of an int for an array of ints),
#   not byte (if you want it to be byte, just make the array argument have
#   type "char*" or "void*" (GCC extension)).
# - nullok: for pointer-based arguments (str, ptr, arr, arr2d), specifies
#   whether the argument is allowed to be NULL.  This is not the default,
#   because checking for NULL pointers results in an additional overhead.
# - zerocopy: for array arguments (arr, arr2d), avoids copying the data
#   to an intermediate buffer.  Can be used at most once for an "in" and at
#   most once for an "out" argument.
# - userbuf: a special sort of zerocopy argument, where the array address
#   on the server side is provided by the backend function, not by zoid.
#   Only works for 1-D arrays (not for arr2d).
#
# Example:
#     int zoidfs_readlink(const zoidfs_handle_t * handle /* in:ptr */,
#                         char * buffer /* out:arr:size=+1 */,
#                         size_t buffer_length /* in:obj */);
#
# Two more files are also generated per invocation:
# - <header>_defs.h: contains numerical identifiers for parsed functions.
# - <header>_dispatch.c: defines an array used by the server side to match
#   numerical identifiers with their corresponding functions.
#
# LIMITATIONS:
# Return type must be scalar or void.
# Structures with pointer arguments cannot be passed around.
# Array outputs introduce extra restrictions.  All the other results must
#    have sizes that can be calculated before a function is invoked.  For
#    arr2d outputs, that also applies to each array line: their sizes cannot
#    change inside the invoked function, if there are any more non-zero-size
#    array lines to follow.  E.g, if the size array of an arr2d is "inout",
#    and initially holds {10, 10, 10}, then after the function returns, it is
#    allowed to hold {0, 0, 0}, or {5, 0, 0}, or {10, 1, 0}, or {10, 10, 10},
#    but *not* {10, 9, 10} (because the modified array line size is followed
#    by another (non-zero-size) line).  Such restrictions apply to 1-D arrays
#    only if they are followed by another output array.  The reason behind
#    these restrictions is that the placement of arrays in result buffer is
#    fixed before a function is invoked, to avoid additional buffering and
#    copies.


$lineno = 0;
$verbose = 0;

if ($#ARGV != 0)
{
    die "Exactly one command-line argument required: header file name!\n";
}

open(HEADER, $ARGV[0]) || die "Cannot open header file $ARGV[0]!\n";
($header_name) = ($ARGV[0] =~ /([^\/]*)$/);
$header_base_name = $header_name;
$header_base_name =~ s/\.h//;

$header_defs_name = $header_base_name . "_defs.h";

$header_id_name = $header_base_name . "_header_id";
$header_id_name =~ tr/a-z/A-Z/;

while (<HEADER>)
{
    $lineno++;

    chop;

    # Ignore contents until an init tag is encountered.
    if (!$scanning)
    {
	next if (! /\/\*\s*START-ZOID-SCANNER\s+ID\s*=\s*(\d+)\s*INIT\s*=\s*(\S+)\s*FINI\s*=\s*(\S+)\s*\*\//);
	$header_id = $1;
	$init_func = $2;
	$fini_func = $3;

	$scanning = 1;
    }

    # Ignore lines with comments.
    if (/^\s*\/\*/)
    {
	$comment_start = $lineno;
	while (! /\*\/\s*$/)
	{
	    if (!($_ = <HEADER>))
	    {
		die "Line $lineno: Unexpected end of file (while" .
		    " ignoring comment started on line $comment_start )!\n";
	    }
	    $lineno++;
	}
	next;
    }

    # Ignore preprocessor directives.
    next if /^\#/;

    # Ignore empty lines.
    next if /^\s*$/;


    # A list of "loop-global" variables we use:
    # $prototype - complete function prototype (with terminating ;),
    # $return_type - function return type,
    # $name - function name,
    # @arguments - function arguments (with comments removed),
    # $arguments_count - number of arguments,
    # @arg_name - function argument names,
    # @arg_inout - in, out or inout.
    # @arg_type - obj, str, ptr, arr, or arr2d
    # @arg_size - index of the size argument for arrays (first arg is 0).
    # @arg_nullok - if NULL is allowed to be passed.
    # @arg_zerocopy - used only temporarily, copied over to $zerocopy_in
    #     and $zerocopy_out.
    # @arg_userbuf - used only temporarily, copied over to $userbuf_in and
    #     $userbuf_out.
    # @function_names - names of functions processed, in order (global).
    # %function_userbuf - userbuf_in of functions processed (global).
    # @arg_copied, @arg_extracted - temporary, used to track dependencies
    #     in case of array arguments.

    # We are assuming to be at the start of a prototype right now.
    # We'll first read the whole thing in, merging into one line.
    /^\s*(.*\S)\s*$/;
    $prototype = $1;
    $proto_start = $lineno;
    while ($prototype !~ /;$/)
    {
	if (!($_ = <HEADER>))
	{
	    die "Line $lineno: Unexpected end of file (while" .
		" reading prototype started on line $proto_start)!\n";
	}
	/^\s*(.*\S)\s*$/;
	$prototype .= " " . $1;
	$lineno++;
    }
    print "prototype: $prototype\n" if ($verbose);


    # Parse the prototype.
    ($return_type, $name, $arguments) = ($prototype =~
	/^(.*\S)\s+([a-zA-Z_]\w*)\s*\(\s*(.*\S)\s*\)\s*;$/);

    if (!defined($return_type) || !defined($name) || !defined($arguments))
    {
	die "Line $lineno: failed to parse prototype $prototype!\n";
    }

    print "return type: '$return_type'\n" if ($verbose);
    print "name: '$name'\n" if ($verbose);
    print "arguments: '$arguments'\n" if ($verbose);


    # Parse the arguments.
    $arguments_count = 0;
    undef(@arguments);
    undef(@arg_name);
    undef(@arg_inout);
    undef(@arg_type);
    undef(@arg_size);
    undef(@arg_nullok);
    undef(@arg_zerocopy);
    undef(@arg_userbuf);
    $zerocopy_in = $zerocopy_out = $userbuf_in = $userbuf_out = -1;
    if ($arguments ne "" && $arguments ne "void")
    {
	@arguments = split(/\s*,\s*/, $arguments);
	$arguments_count = $#arguments + 1;

	# Parse a single argument.
	$i = 0;
	foreach $argument (@arguments)
	{
	    print "argument: '$argument'\n" if ($verbose);

	    # Split into C argument and comment.
	    if ($argument =~ /^(.*\S)\s*\/\*\s*(.*\S)\s*\*\//)
	    {
		$argument = $1;
		$comment = $2;
	    }
	    else
	    {
		undef $comment;
	    }

	    if ($verbose && defined($comment))
	    {
		print "argument part: '$argument'\n";
		print "comment part: '$comment'\n";
	    }

	    # Extract argument name.
	    ($arg_name) =
		($argument =~ /([a-zA-Z_]\w*)\s*\)*\s*(\[\s*\]){0,1}$/);
	    if (!defined($arg_name))
	    {
		die "Line $lineno: failed to extract argument name" .
		    " from $argument!\n";
	    }
	    print "argument name: '$arg_name'\n" if ($verbose);

	    # Put back the argument without comment.
	    $arguments[$i] = $argument;

	    $arg_name[$i] = $arg_name;

	    # Put in the default attributes.
	    $arg_inout[$i] = "in";
	    $arg_type[$i] = "obj";

	    # Parse the comment part.
	    if (defined($comment))
	    {
		@attrs = split(/\s*:\s*/, $comment);

		foreach $attr (@attrs)
		{
		    if ($attr eq "in" || $attr eq "out" || $attr eq "inout")
		    {
			$arg_inout[$i] = $attr;
		    }
		    else
		    {
			if ($attr eq "obj" || $attr eq "str" ||
			    $attr eq "ptr" || $attr eq "arr" ||
			    $attr eq "arr2d")
			{
			    $arg_type[$i] = $attr;
			}
			else
			{
			    if ($attr =~ /^size\s*=\s*([+-]{0,1})(\d+)$/)
			    {
				if ($1 eq "")
				{
				    $size = $2 - 1;
				}
				else
				{
				    $size = $i  - $2 if ($1 eq "-");
				    $size = $i  + $2 if ($1 eq "+");
				}
				if ($size < 0 || $size >= $arguments_count)
				{
				    die "Line $lineno: invalid value in " .
					"$attr!\n";
				}
				$arg_size[$i] = $size;
			    }
			    else
			    {
				if ($attr eq "nullok")
				{
				    $arg_nullok[$i] = 1;
				}
				else
				{
				    if ($attr eq "zerocopy")
				    {
					$arg_zerocopy[$i] = 1;
				    }
				    else
				    {
					if ($attr eq "userbuf")
					{
					    $arg_userbuf[$i] = 1;
					}
					else
					{
					    die "Line $lineno: unknown " .
						"attribute $attr!\n";
					}
				    }
				}
			    }
			}
		    }
		}

		if ($verbose)
		{
		    print "I/O: $arg_inout[$i], Type: $arg_type[$i]";
		    print ", Size: $arg_size[$i]" if (defined($arg_size[$i]));
		    print "\n";
		}

		# Check for invalid combinations.
		if (($arg_inout[$i] =~ /out$/ && ($arg_type[$i] eq "obj" ||
						  $arg_type[$i] eq "str")) ||
		    (defined($arg_size[$i]) && $arg_type[$i] !~ /^arr/) ||
		    ($arg_type[$i] =~ /^arr/ && !defined($arg_size[$i])) ||
		    ($arg_nullok[$i] && $arg_type[$i] eq "obj") ||
		    ($arg_zerocopy[$i] && $arg_type[$i] !~ /^arr/) ||
		    ($arg_userbuf[$i] && ($arg_type[$i] ne "arr" ||
					  !defined($arg_zerocopy[$i]))))
		{
		    die "Line $lineno: invalid attribute " .
			"combination $comment!\n";
		}
		if ($arg_zerocopy[$i])
		{
		    if ($arg_inout[$i] =~ /^in/)
		    {
			if ($zerocopy_in != -1)
			{
			    die "Line $lineno: duplicate input zero-copy!\n";
			}
			$zerocopy_in = $i;
		    }
		    if ($arg_inout[$i] =~ /out$/)
		    {
			if ($zerocopy_out != -1)
			{
			    die "Line $lineno: duplicate output zero-copy!\n";
			}
			$zerocopy_out = $i;
		    }
		}
		if ($arg_userbuf[$i])
		{
		    if ($arg_inout[$i] =~ /^in/)
		    {
			if ($userbuf_in != -1)
			{
			    die "Line $lineno: duplicate input userbuf!\n";
			}
			$userbuf_in = $i;
		    }
		    if ($arg_inout[$i] =~ /out$/)
		    {
			if ($userbuf_out != -1)
			{
			    die "Line $lineno: duplicate output userbuf!\n";
			}
			$userbuf_out = $i;
		    }
		}
	    }

	    $i++;
	}
    }

    push(@function_names, $name);
    $function_userbuf{$name} = ($userbuf_in != -1);

    # Input is parsed.  Time to generate the output.

    # First the client stubs.
    open(CLIENT_OUT, ">client/$name.c") ||
	die "Can't open client/$name.c: $!\n";

    # Output function header.
    print CLIENT_OUT "/* Automatically generated.  Do not edit!  */\n\n" .
	"#include \"$header_name\"\n\n" .
	"#include <errno.h>\n" .
	"#include <stdlib.h>\n" .
	"#include <string.h>\n\n" .
	"#include \"zoid_api.h\"\n\n" .
	"#include \"$header_defs_name\"\n\n";
    print CLIENT_OUT "$return_type\n$name(";
    if ($arguments_count > 0)
    {
	print CLIENT_OUT $arguments[0];
	$lead = " " x (length($name) + 1);
	$lead =~ s/        /\t/g;
	foreach $argument (@arguments[1..$#arguments])
	{
	    print CLIENT_OUT ",\n" .
		"$lead$argument";
	}
    }
    else
    {
	print CLIENT_OUT "void";
    }
    print CLIENT_OUT ")\n{\n";

    # Output local variables.
    print CLIENT_OUT "    int command_id = \U${name}\E_ID;\n" .
	"    char *zoid_buffer, *current_cmdbuf, *current_resbuf;\n";
    print CLIENT_OUT "    int cmd_needed, res_needed;\n";

    if ($return_type ne "void")
    {
	print CLIENT_OUT "    $return_type return_value;\n";
    }

    # Calculate the space needed for arguments.
    print CLIENT_OUT "\n" .
	"    cmd_needed = sizeof(command_id);\n";
    undef(@arg_copied);
    # For input userbuf, we ensure that size is always the first argument.
    calculate_argument_size($arg_size[$userbuf_in]) if ($userbuf_in != -1);
    for ($i = 0; $i < $arguments_count; $i++)
    {
	if (($arg_inout[$i] =~ /^in/ || $arg_nullok[$i]) &&
	    !defined($arg_copied[$i]) && $i != $zerocopy_in)
	{
	    calculate_argument_size($i);
	}
    }
    calculate_argument_size($zerocopy_in) if ($zerocopy_in != -1);

    # Calculate the space needed for results.
    print CLIENT_OUT "\n" .
	"    res_needed = 0;\n";
    if ($zerocopy_out != -1)
    {
	# Space for the offset to zero-copy data.
	print CLIENT_OUT
	    "    res_needed += __zoid_0c_out_buf_hdr1;\n";
    }
    if ($return_type ne "void")
    {
	print CLIENT_OUT
	    "    res_needed += sizeof(return_value);\n";
    }
    undef(@arg_copied);
    for ($i = 0; $i < $arguments_count; $i++)
    {
	if ($arg_inout[$i] =~ /out$/ && !defined($arg_copied[$i]) &&
	    $i != $zerocopy_out)
	{
	    calculate_output_size(\*CLIENT_OUT, $i);
	}
    }
    if ($zerocopy_out != -1)
    {
	calculate_output_size(\*CLIENT_OUT, $zerocopy_out);
    }

    # Provide a buffer for command and results.
    print CLIENT_OUT "\n" .
	"    if (!(zoid_buffer = __zoid_allocate_buffer(" .
	"cmd_needed > res_needed ?\n" .
	"\t\t\t\t\t       cmd_needed : res_needed)))\n" .
	"\treturn" . ($return_type eq "void" ? "" : " -1") . ";\n";

    print CLIENT_OUT "\n" .
	"    current_cmdbuf = zoid_buffer;\n";

    # Copy command ID.
    print CLIENT_OUT "\n" .
	"    memcpy(current_cmdbuf, &command_id, sizeof(command_id));\n" .
	"    current_cmdbuf += sizeof(command_id);\n";

    # Copy the arguments.
    undef(@arg_copied);
    copy_argument($arg_size[$userbuf_in]) if ($userbuf_in != -1);
    for ($i = 0; $i < $arguments_count; $i++)
    {
	if (($arg_inout[$i] =~ /^in/ || $arg_nullok[$i]) &&
	    !defined($arg_copied[$i]))
	{
	    copy_argument($i);
	}
    }

    # Output function invocation.
    print CLIENT_OUT "\n" .
	"    if (__zoid_submit_command(zoid_buffer, cmd_needed, res_needed, " .
	($userbuf_in != -1 ? "1" : "0") . ",\n";

    if ($zerocopy_in != -1)
    {
	if ($arg_type[$zerocopy_in] eq "arr2d")
	{
	    print CLIENT_OUT "\t$arg_name[$zerocopy_in], " .
		"$arg_name[$arg_size[$zerocopy_in]], " .
		($arg_type[$arg_size[$arg_size[$zerocopy_in]]] eq "ptr" ?
		 "*" : "") .
		"$arg_name[$arg_size[$arg_size[$zerocopy_in]]], " .
		"sizeof(**$arg_name[$zerocopy_in]),\n";
	}
	else
	{
	    print CLIENT_OUT "\tNULL, $arg_name[$zerocopy_in], " .
		($arg_type[$arg_size[$zerocopy_in]] eq "ptr" ? "*" : "") .
		"$arg_name[$arg_size[$zerocopy_in]], " .
		"sizeof(*$arg_name[$zerocopy_in]),\n";
	}
    }
    else
    {
	print CLIENT_OUT "\tNULL, NULL, 0, 0,\n";
    }

    if ($zerocopy_out != -1)
    {
	print CLIENT_OUT "\t$arg_name[$zerocopy_out]";
    }
    else
    {
	print CLIENT_OUT "\tNULL";
    }

    print CLIENT_OUT ") == -1)\n" .
	"    {\n" .
	"\t__zoid_release_buffer(zoid_buffer);\n" .
	"\treturn" . ($return_type eq "void" ? "" : " -1") . ";\n" .
	"    }\n";

    # Extract the results.
    print CLIENT_OUT "\n" .
	"    current_resbuf = zoid_buffer;\n";

    if ($zerocopy_out != -1)
    {
	# Skip the offset to zerocopy data.
	print CLIENT_OUT "\n" .
	    "    current_resbuf += __zoid_0c_out_buf_hdr1;\n";
    }

    # Extract return value.
    if ($return_type ne "void")
    {
	print CLIENT_OUT "\n" .
	    "    memcpy(&return_value, current_resbuf, " .
	    "sizeof(return_value));\n" .
	    "    current_resbuf += sizeof(return_value);\n";
    }

    # Extract output.
    undef(@arg_extracted);
    for ($i = 0; $i < $arguments_count; $i++)
    {
	if ($arg_inout[$i] =~ /out$/ && !defined($arg_extracted[$i]))
	{
	    extract_output($i);
	}
    }

    # Release the output buffer.
    print CLIENT_OUT "\n" .
	"    __zoid_release_buffer(zoid_buffer);\n";

    # Return result to the caller.
    if ($return_type ne "void")
    {
	print CLIENT_OUT "\n" .
	    "    return return_value;\n";
    }

    print CLIENT_OUT "}\n";

    close(CLIENT_OUT);

    # Client stubs are done.  Now generate the corresponding server side.
    open(SERVER_OUT, ">server/$name.c") ||
	die "Can't open server/$name.c: $!\n";

    # Output function header.
    print SERVER_OUT "/* Automatically generated.  Do not edit!  */\n\n" .
	"#include \"$header_name\"\n\n" .
	"#include <errno.h>\n" .
	"#include <stdlib.h>\n" .
	"#include <string.h>\n\n" .
	"#include \"zoid_api.h\"\n\n";
    print SERVER_OUT "char*\n__zoid_invoke_$name(char* command_buffer, " .
	"int* result_len, int* total_len";
    if ($userbuf_in != -1)
    {
	$lead = " " x length("__zoid_invoke_$name(");
	$lead =~ s/        /\t/g;
	print SERVER_OUT ",\n" .
	    "$lead" . "void* user_buffer";
    }
    print SERVER_OUT ")\n" .
	"{\n" .
	"    char *current_cmdbuf, *result_buffer, *current_resbuf;\n" .
	"    int res_needed = 0;\n";

    if ($return_type ne "void")
    {
	print SERVER_OUT "    $return_type return_value;\n";
    }

    # Define local variables.
    for ($i = 0; $i < $arguments_count; $i++)
    {
	define_var($i);
    }

    print SERVER_OUT "\n" .
	"    current_cmdbuf = command_buffer;\n";

    # Extract arguments.
    undef(@arg_extracted);
    extract_argument($arg_size[$userbuf_in]) if ($userbuf_in != -1);
    for ($i = 0; $i < $arguments_count; $i++)
    {
	if (($arg_inout[$i] =~ /^in/ || $arg_nullok[$i]) &&
	    !defined($arg_extracted[$i]) && $i != $zerocopy_in)
	{
	    extract_argument($i);
	}
    }
    extract_argument($zerocopy_in) if ($zerocopy_in != -1);

    # Calculate the space needed for results.
    print SERVER_OUT "\n";

    if ($zerocopy_out != -1)
    {
	# Allocate space for the offset to zero-copy data.
	print SERVER_OUT
	    "    res_needed += __zoid_0c_out_buf_hdr1;\n";
    }
    if ($return_type ne "void")
    {
	print SERVER_OUT
	    "    res_needed += sizeof(return_value);\n";
    }
    undef(@arg_copied);
    for ($i = 0; $i < $arguments_count; $i++)
    {
	if ($arg_inout[$i] =~ /out$/ && !defined($arg_copied[$i]) &&
	    $i != $zerocopy_out)
	{
	    calculate_output_size(\*SERVER_OUT, $i);
	}
    }
    if ($zerocopy_out != -1)
    {
	calculate_output_size(\*SERVER_OUT, $zerocopy_out);
    }

    # Determine the result buffer.
    print SERVER_OUT "\n" .
	"    if (!(result_buffer = __zoid_allocate_buffer(res_needed)))\n" .
	"\treturn NULL;\n";

    # Make adjustments needed for output arrays.
    for ($i = 0; $i < $arguments_count; $i++)
    {
	prepare_result_buf($i);
    }

    # Invoke the function.
    print SERVER_OUT "\n    ";
    print SERVER_OUT "return_value = " if ($return_type ne "void");
    print SERVER_OUT "$name(";
    for ($i = 0; $i < $arguments_count; $i++)
    {
	print SERVER_OUT ", " if ($i > 0);
	print SERVER_OUT $arg_name[$i];
    }
    print SERVER_OUT ");\n";

    print SERVER_OUT "\n    current_resbuf = result_buffer;\n";

    if ($zerocopy_out != -1)
    {
	# Allocate space for the offset to zero-copy data.
	print SERVER_OUT "\n" .
	    "    current_resbuf += __zoid_0c_out_buf_hdr1;\n";
    }

    # Copy the return value.
    if ($return_type ne "void")
    {
	print SERVER_OUT "\n" .
	    "    memcpy(current_resbuf, &return_value, " .
	    "sizeof(return_value));\n" .
	    "    current_resbuf += sizeof(return_value);\n";
    }

    # Copy other results.
    undef(@arg_copied);
    for ($i = 0; $i < $arguments_count; $i++)
    {
	if ($arg_inout[$i] =~ /out$/ && !defined($arg_copied[$i]) &&
	    $i != $zerocopy_out)
	{
	    copy_output($i);
	}
    }
    print SERVER_OUT "\n" .
	"    *result_len = current_resbuf - result_buffer;\n" .
	"    *total_len = *result_len;\n";
    if ($zerocopy_out != -1)
    {
	copy_output($zerocopy_out);
    }

    # Finish.
    print SERVER_OUT "\n" .
	"    return result_buffer;\n" .
	"}\n";

    close(SERVER_OUT);

    print "\n" if ($verbose);
}

if (!$scanning)
{
    die "Did not encounter a valid START-ZOID-SCANNER directive!\n";
}

# We have parsed the whole input file and generated stubs for individual
# functions, both client and server side.  We still need to generate some
# server-side management files.

$header_defs_name_h = $header_defs_name;
$header_defs_name_h =~ tr/./_/;

open(SERVER_HEADER, ">$header_defs_name") ||
    die "Cannot open $header_defs_name!\n";

print SERVER_HEADER "/* Automatically generated.  Do not edit!  */\n\n" .
    "#ifndef \U${header_defs_name_h}\E\n" .
    "#define \U${header_defs_name_h}\E\n\n";

print SERVER_HEADER "#define $header_id_name $header_id\n\n";

$id_count = 0;
foreach $name (@function_names)
{
    print SERVER_HEADER "#define \U${name}\E_ID " .
	"($header_id_name << 16 | $id_count)\n";
    $id_count++;
}
print SERVER_HEADER "\n";
foreach $name (@function_names)
{
    print SERVER_HEADER "char* __zoid_invoke_$name(char* current_cmdbuf, " .
	"int* result_len, int* total_len" .
	($function_userbuf{$name} == 1 ? ", void* user_buffer" : "") .
	");\n";
    if ($function_userbuf{$name} == 1)
    {
	print SERVER_HEADER "void* ${name}_allocate_cb(int buffer_len);\n";
    }
}

print SERVER_HEADER "\n#endif\n";

close(SERVER_HEADER);

$server_dispatch_name = $header_base_name . "_dispatch.c";

open(SERVER_DISPATCH, ">$server_dispatch_name") ||
    die "Cannot open $server_dispatch_name!\n";

print SERVER_DISPATCH "/* Automatically generated.  Do not edit!  */\n\n" .
    "#include \"zoid_api.h\"\n\n" .
    "#include \"$header_defs_name\"\n\n" .
    "static struct dispatch_array dispatch_array[] =\n" .
    "{\n";

foreach $name (@function_names)
{
    print SERVER_DISPATCH "    {__zoid_invoke_${name}, " .
	($function_userbuf{$name} ? "${name}_allocate_cb" : "NULL") .
	"},\n";
}


print SERVER_DISPATCH "};\n\n";

print SERVER_DISPATCH "void $init_func(int);\n\n" if ($init_func ne "NULL");
print SERVER_DISPATCH "void $fini_func(void);\n\n" if ($fini_func ne "NULL");

print SERVER_DISPATCH "void " . $header_base_name .
    "__register_functions(void) __attribute__((constructor));\n\n";

print SERVER_DISPATCH "void " . $header_base_name .
    "__register_functions(void)\n" .
    "{\n" .
    "    __zoid_register_functions($header_id_name, dispatch_array,\n" .
    "\t\t\t      sizeof(dispatch_array) / sizeof(dispatch_array[0]),\n" .
    "\t\t\t      $init_func, $fini_func);\n" .
    "}\n";

close(SERVER_DISPATCH);

exit 0;


# Used when outputting client stub.  Outputs a code to calculate the argument
# size.
sub calculate_argument_size
{
    my $arg = $_[0];

    # For arrays, we need to be sure that size precedes them.
    if (defined($arg_size[$arg]) && $arg_size[$arg] > $arg &&
	!defined($arg_copied[$arg_size[$arg]]))
    {
	calculate_argument_size($arg_size[$arg]);
    }

    # For arguments that are allowed to be NULL, we must send an indicator
    # whether they are.  We have to do it even for arguments that are only
    # output arguments -- in that case we return prematurely.
    # We also don't count data if it concerns a zero-copy argument, since
    # it doesn't get copied to the command buffer.
    if ($arg_nullok[$arg])
    {
	print CLIENT_OUT 
	    "    cmd_needed++;\n";
	return if ($arg_inout[$arg] !~ /^in/ || $arg == $zerocopy_in);
	print CLIENT_OUT "    if ($arg_name[$arg])\n" .
	    "    {\n";
    }
    else
    {
	return if ($arg == $zerocopy_in);
    }

    if ($arg_type[$arg] eq "obj")
    {
	print CLIENT_OUT
	    "    cmd_needed += sizeof($arg_name[$arg]);\n";
    }
    if ($arg_type[$arg] eq "ptr")
    {
	print CLIENT_OUT
	    "    cmd_needed += sizeof(*$arg_name[$arg]);\n";
    }
    if ($arg_type[$arg] eq "str")
    {
	print CLIENT_OUT
	    "    cmd_needed += strlen($arg_name[$arg]) + 1;\n";
    }
    if ($arg_type[$arg] eq "arr")
    {
	print CLIENT_OUT
	    "    cmd_needed += sizeof(*$arg_name[$arg]) * " .
	    ($arg_type[$arg_size[$arg]] eq "ptr" ? "*" : "") .
	    "$arg_name[$arg_size[$arg]];\n";
    }
    if ($arg_type[$arg] eq "arr2d")
    {
	print CLIENT_OUT "\n" .
	    "    {\n" .
	    "\tint i;\n" .
	    "\tfor (i = 0; i < " .
	    ($arg_type[$arg_size[$arg_size[$arg]]] eq "ptr" ? "*" : "") .
	    "$arg_name[$arg_size[$arg_size[$arg]]]; i++)\n" .
	    "\t{\n" .
	    "\t    cmd_needed += sizeof(**$arg_name[$arg]) * " .
	    "$arg_name[$arg_size[$arg]]"."[i];\n" .
	    "\t}\n" .
	    "    }\n";
    }

    print CLIENT_OUT "    }\n" if ($arg_nullok[$arg]);

    $arg_copied[$arg] = 1;
}

# Used when outputting client stub.  Outputs an argument-to-memory copy
# code for a single argument.
sub copy_argument
{
    my $arg = $_[0];

    # Don't copy a zero-copy argument.
    return if ($arg == $zerocopy_in);

    # For arrays, we need to be sure that size precedes them.
    if (defined($arg_size[$arg]) && $arg_size[$arg] > $arg &&
	!defined($arg_copied[$arg_size[$arg]]))
    {
	copy_argument($arg_size[$arg]);
    }

    # For arguments that are allowed to be NULL, we must send an indicator
    # whether they are.  We have to do it even for arguments that are only
    # output arguments -- in that case we return prematurely.
    if ($arg_nullok[$arg])
    {
	print CLIENT_OUT "\n" .
	    "    *current_cmdbuf++ = !!$arg_name[$arg];\n";
	return if ($arg_inout[$arg] !~ /^in/);
	print CLIENT_OUT "    if ($arg_name[$arg])\n" .
	    "    {";
    }

    if ($arg_type[$arg] eq "obj")
    {
	print CLIENT_OUT "\n" .
	    "    memcpy(current_cmdbuf, &$arg_name[$arg], " .
	    "sizeof($arg_name[$arg]));\n" .
	    "    current_cmdbuf += sizeof($arg_name[$arg]);\n";
    }
    if ($arg_type[$arg] eq "ptr")
    {
	print CLIENT_OUT "\n" .
	    "    memcpy(current_cmdbuf, $arg_name[$arg], " .
	    "sizeof(*$arg_name[$arg]));\n" .
	    "    current_cmdbuf += sizeof(*$arg_name[$arg]);\n";
    }
    if ($arg_type[$arg] eq "str")
    {
	print CLIENT_OUT "\n" .
	    "    strcpy(current_cmdbuf, $arg_name[$arg]);\n" .
	    "    current_cmdbuf += strlen($arg_name[$arg]) + 1;\n";
    }
    if ($arg_type[$arg] eq "arr")
    {
	print CLIENT_OUT "\n" .
	    "    memcpy(current_cmdbuf, $arg_name[$arg], " .
	    "sizeof(*$arg_name[$arg]) * " .
	    ($arg_type[$arg_size[$arg]] eq "ptr" ? "*" : "") .
	    "$arg_name[$arg_size[$arg]]);\n" .
	    "    current_cmdbuf += sizeof(*$arg_name[$arg]) * " .
	    ($arg_type[$arg_size[$arg]] eq "ptr" ? "*" : "") .
	    "$arg_name[$arg_size[$arg]];\n";
    }
    if ($arg_type[$arg] eq "arr2d")
    {
	print CLIENT_OUT "\n" .
	    "    {\n" .
	    "\tint i;\n" .
	    "\tfor (i = 0; i < " .
	    ($arg_type[$arg_size[$arg_size[$arg]]] eq "ptr" ? "*" : "") .
	    "$arg_name[$arg_size[$arg_size[$arg]]]; i++)\n" .
	    "\t{\n" .
	    "\t    memcpy(current_cmdbuf, $arg_name[$arg]"."[i], " .
	    "sizeof(**$arg_name[$arg]) * " .
	    "$arg_name[$arg_size[$arg]]"."[i]);\n" .
	    "\t    current_cmdbuf += sizeof(**$arg_name[$arg]) * " .
	    "$arg_name[$arg_size[$arg]]"."[i];\n" .
	    "\t}\n" .
	    "    }\n";
    }

    print CLIENT_OUT "    }\n" if ($arg_nullok[$arg]);

    $arg_copied[$arg] = 1;
}

# Used when outputting client stub.  Outputs a memory-to-result copy
# code for a single result.
sub extract_output
{
    my $arg = $_[0];

    # Don't copy a zero-copy result.
    return if ($arg == $zerocopy_out);

    # For arrays, we need to be sure that out size, if any, precedes them.
    if (defined($arg_size[$arg]) && $arg_size[$arg] > $arg &&
	$arg_inout[$arg_size[$arg]] =~ /out$/ &&
	!defined($arg_extracted[$arg_size[$arg]]))
    {
	extract_output($arg_size[$arg]);
    }

    if ($arg_nullok[$arg])
    {
	print CLIENT_OUT "    if ($arg_name[$arg])\n" .
	    "    {";
    }

    if ($arg_type[$arg] eq "ptr")
    {
	print CLIENT_OUT "\n" .
	    "    memcpy($arg_name[$arg], current_resbuf, " .
	    "sizeof(*$arg_name[$arg]));\n" .
	    "    current_resbuf += sizeof(*$arg_name[$arg]);\n";
    }
    if ($arg_type[$arg] eq "arr")
    {
	print CLIENT_OUT "\n" .
	    "    memcpy($arg_name[$arg], current_resbuf, " .
	    "sizeof(*$arg_name[$arg]) * " .
	    ($arg_type[$arg_size[$arg]] eq "ptr" ? "*" : "") .
	    "$arg_name[$arg_size[$arg]]);\n" .
	    "    current_resbuf += sizeof(*$arg_name[$arg]) * " .
	    ($arg_type[$arg_size[$arg]] eq "ptr" ? "*" : "") .
	    "$arg_name[$arg_size[$arg]];\n";
    }
    if ($arg_type[$arg] eq "arr2d")
    {
	print CLIENT_OUT "\n" .
	    "    {\n" .
	    "\tint i;\n" .
	    "\tfor (i = 0; i < " .
	    ($arg_type[$arg_size[$arg_size[$arg]]] eq "ptr" ? "*" : "") .
	    "$arg_name[$arg_size[$arg_size[$arg]]]; i++)\n" .
	    "\t{\n" .
	    "\t    memcpy($arg_name[$arg]"."[i], current_resbuf, " .
	    "sizeof(**$arg_name[$arg]) * $arg_name[$arg_size[$arg]]" .
	    "[i]);\n" .
	    "\t    current_resbuf += sizeof(**$arg_name[$arg]) * " .
	    "$arg_name[$arg_size[$arg]]"."[i];\n" .
	    "\t}\n" .
	    "    }\n";
    }

    print CLIENT_OUT "    }\n" if ($arg_nullok[$arg]);

    $arg_extracted[$arg] = 1;
}

# Used when outputting server code.  Creates variables needed for
# a particular argument.
sub define_var
{
    my $arg = $_[0];
    my $var = $arguments[$arg];

    # Replace array notation with an asterisk for arrays.
    if ($var =~ /\[\s*\]$/)
    {
	$var =~ s/([a-zA-Z_]\w*)\s*(\)*)\s*\[\s*\]$/\*$1$2/;
    }

    if ($arg_type[$arg] eq "ptr")
    {
	# Remove one level of const.
	$var =~ s/^const\s*//;

	# Remove one level of pointers.  Also add a "_t" suffix to
	# variable name.
	my $var_t = $var;
	if (!($var_t =~ s/\*\s*([a-zA-Z_]\w*)\s*(\)*)$/$1_t$2/))
	{
	    die "Line $lineno: could not extract the base type for pointer " .
		"argument $arg!\n";
	}
	my $var_t_name;
	($var_t_name) = ($var_t =~ /\s*([a-zA-Z_]\w*)\s*\)*$/);

	print SERVER_OUT "    $var_t;\n";
	print SERVER_OUT "    $var = &$var_t_name;\n";
    }
    else
    {
	print SERVER_OUT "    $var;\n";
    }
    # Define a variable to hold an offset to the result buffer.
    if ($arg_type[$arg] =~ /^arr/ && $arg_inout[$arg] =~ /out$/ &&
	$arg != $userbuf_out)
    {
	print SERVER_OUT "    int $arg_name[$arg]_offset";
	# We 0-initialize just to avoid a GCC warning...
	print SERVER_OUT " = 0" if ($arg_nullok[$arg]);
	print SERVER_OUT ";\n";
    }

    return $arg;
}

# Used when outputting server code.  Outputs a memory-to-variable copy
# code for a single argument.
sub extract_argument
{
    my $arg = $_[0];

    # For arrays, we need to be sure that size precedes them.
    if (defined($arg_size[$arg]) && $arg_size[$arg] > $arg &&
	!defined($arg_extracted[$arg_size[$arg]]))
    {
	extract_argument($arg_size[$arg]);
    }

    if ($arg_nullok[$arg])
    {
	print SERVER_OUT "\n" .
	    "    if (!*current_cmdbuf++)\n" .
	    "\t$arg_name[$arg] = NULL;\n";
	if ($arg_inout[$arg] !~ /^in/)
	{
	    # This is nullok out-only variable.  If it is an array, then the
	    # pointer is uninitialized if it is not NULL.  We have to
	    # initialize it to something -- anything non-NULL, since we
	    # will overwrite it later on, but only if the variable is non-NULL.
	    if ($arg_type[$arg] =~ /^arr/)
	    {
		print SERVER_OUT "    else\n" .
		    "\t$arg_name[$arg] = (void*)current_cmdbuf;\n";
	    }
	    return;
	}
	print SERVER_OUT "    else\n" .
	    "    {";
    }

    if ($arg_type[$arg] eq "obj")
    {
	print SERVER_OUT "\n" .
	    "    memcpy(&$arg_name[$arg], current_cmdbuf, " .
	    "sizeof($arg_name[$arg]));\n" .
	    "    current_cmdbuf += sizeof($arg_name[$arg]);\n";
    }
    if ($arg_type[$arg] eq "ptr")
    {
	print SERVER_OUT "\n" .
	    "    memcpy($arg_name[$arg], current_cmdbuf, " .
	    "sizeof(*$arg_name[$arg]));\n" .
	    "    current_cmdbuf += sizeof(*$arg_name[$arg]);\n";
    }
    if ($arg_type[$arg] eq "str")
    {
	print SERVER_OUT "\n" .
	    "    $arg_name[$arg] = current_cmdbuf;\n" .
	    "    current_cmdbuf += strlen($arg_name[$arg]) + 1;\n";
    }
    if ($arg_type[$arg] eq "arr")
    {
	print SERVER_OUT "\n";
	if ($arg != $userbuf_in)
	{
	    if ($arg == $zerocopy_in)
	    {
		print SERVER_OUT
		    "    current_cmdbuf = __zoid_0c_in_buf(command_buffer, " .
		    "current_cmdbuf);\n";
	    }

	    print SERVER_OUT
		"    $arg_name[$arg] = (void*)current_cmdbuf;\n" .
		"    current_cmdbuf += sizeof(*$arg_name[$arg]) * " .
		($arg_type[$arg_size[$arg]] eq "ptr" ? "*" : "") .
		"$arg_name[$arg_size[$arg]];\n";
	}
	else
	{
	    print SERVER_OUT
		"    $arg_name[$arg] = user_buffer;\n";
	}
    }
    if ($arg_type[$arg] eq "arr2d")
    {
	# The only sort-of dynamic allocation directly in the stubs...
	print SERVER_OUT "\n" .
	    "    $arg_name[$arg] = alloca(sizeof(*$arg_name[$arg]) * " .
	    ($arg_type[$arg_size[$arg_size[$arg]]] eq "ptr" ? "*" : "") .
	    "$arg_name[$arg_size[$arg_size[$arg]]]);\n" .
	    "    {\n" .
	    "\tint i;\n" .
	    "\tfor (i = 0; i < " .
	    ($arg_type[$arg_size[$arg_size[$arg]]] eq "ptr" ? "*" : "") .
	    "$arg_name[$arg_size[$arg_size[$arg]]]; i++)\n" .
	    "\t{\n";

	if ($arg == $zerocopy_in)
	{
	    print SERVER_OUT
		"\t    current_cmdbuf = __zoid_0c_in_buf(" .
		"command_buffer, current_cmdbuf);\n";
	}

	print SERVER_OUT
	    "\t    $arg_name[$arg]"."[i] = (void*)current_cmdbuf;\n" .
	    "\t    current_cmdbuf += sizeof(**$arg_name[$arg]) * " .
	    "$arg_name[$arg_size[$arg]]"."[i];\n" .
	    "\t}\n" .
	    "    }\n";
    }

    print SERVER_OUT "    }\n" if ($arg_nullok[$arg]);

    $arg_extracted[$arg] = 1;
}

# Used when outputting server code.  Outputs a code to calculate the buffer
# size needed for results.
sub calculate_output_size
{
    my $fh = $_[0];
    my $arg = $_[1];

    # For arrays, we need to be sure that out size, if any, precedes them.
    if (defined($arg_size[$arg]) && $arg_size[$arg] > $arg &&
	$arg_inout[$arg_size[$arg]] =~ /out$/ &&
	!defined($arg_copied[$arg_size[$arg]]))
    {
	calculate_output_size($fh, $arg_size[$arg]);
    }

    if ($arg_nullok[$arg])
    {
	print $fh "    if ($arg_name[$arg])\n" .
	    "    {\n";
    }

    if ($arg_type[$arg] eq "ptr")
    {
	print $fh
	    "    res_needed += sizeof(*$arg_name[$arg]);\n";
    }
    if ($arg == $zerocopy_out)
    {
	# We need space for the major dimension size and the size in bytes
	# of the first line of minor dimension.
	print $fh
	    "    res_needed += __zoid_0c_out_buf_hdr2;\n";
    }

    # We don't include the size of the zero-copy data on the client side,
    # but we do include it on the server, since it's part of the same buffer.
    # Even on the server, we don't copy it if it's a user-provided buffer.
    if ($arg != $zerocopy_out || ($fh == \*SERVER_OUT && $arg != $userbuf_out))
    {
	if ($arg_type[$arg] eq "arr")
	{
	    if ($arg == $zerocopy_out)
	    {
		print $fh
		    "    res_needed = __zoid_0c_buf_align(res_needed);\n";
	    }

	    if ($fh == \*SERVER_OUT)
	    {
		print $fh
		    "    $arg_name[$arg]_offset = res_needed;\n";
	    }

	    print $fh
		"    res_needed += sizeof(*$arg_name[$arg]) * " .
		($arg_type[$arg_size[$arg]] eq "ptr" ? "*" : "") .
		"$arg_name[$arg_size[$arg]];\n";
	}
	if ($arg_type[$arg] eq "arr2d")
	{
	    print $fh
		"    {\n" .
		"\tint i;\n";
	    if ($fh == \*SERVER_OUT)
	    {
		print $fh
		    "\t$arg_name[$arg]_offset = res_needed;\n";
	    }
	    print $fh
		"\tfor (i = 0; i < " .
		($arg_type[$arg_size[$arg_size[$arg]]] eq "ptr" ? "*" : "") .
		"$arg_name[$arg_size[$arg_size[$arg]]]; i++)\n" .
		"\t{\n";
	    if ($arg == $zerocopy_out)
	    {
		print $fh
		    "\t    res_needed = __zoid_0c_buf_align(res_needed);\n";
	    }
	    print $fh
		"\t    res_needed += sizeof(**$arg_name[$arg]) * " .
		"$arg_name[$arg_size[$arg]]"."[i];\n";
	    if ($arg == $zerocopy_out)
	    {
		# Allocate space for the integer specifying the size of
		# the next entry.

		print $fh
		    "\t    res_needed += __zoid_0c_out_buf_ftr(i, " .
		    "$arg_name[$arg_size[$arg_size[$arg]]]);\n";
	    }
	    print $fh
		"\t}\n" .
		"    }\n";
	}
    }

    print $fh "    }\n" if ($arg_nullok[$arg]);

    $arg_copied[$arg] = 1;
}

# Used when outputting server code.  Outputs a result-to-memory copy
# code for a single result.
sub copy_output
{
    my $arg = $_[0];

    # For arrays, we need to be sure that out size, if any, precedes them.
    if (defined($arg_size[$arg]) && $arg_size[$arg] > $arg &&
	$arg_inout[$arg_size[$arg]] =~ /out$/ &&
	!defined($arg_copied[$arg_size[$arg]]))
    {
	copy_output($arg_size[$arg]);
    }

    if ($arg_nullok[$arg])
    {
	print SERVER_OUT "    if ($arg_name[$arg])\n" .
	    "    {";
    }

    if ($arg_type[$arg] eq "ptr")
    {
	print SERVER_OUT "\n" .
	    "    memcpy(current_resbuf, $arg_name[$arg], " .
	    "sizeof(*$arg_name[$arg]));\n" .
	    "    current_resbuf += sizeof(*$arg_name[$arg]);\n";
    }

    if ($arg == $zerocopy_out)
    {
	print SERVER_OUT "\n" .
	    "    *result_len += __zoid_0c_out_buf_hdr2;\n" .
	    "    *total_len += __zoid_0c_out_buf_hdr2;\n" .
	    "    {\n";
	print SERVER_OUT "\tint i;\n" if ($arg_type[$arg] eq "arr2d");
	print SERVER_OUT
	    "\tint tmp = current_resbuf - result_buffer;\n" .
	    "\tmemcpy(result_buffer, &tmp, sizeof(tmp));\n";
	if ($arg_type[$arg] eq "arr")
	{
	    print SERVER_OUT
		"\ttmp = -1;\n" .
		"\tmemcpy(current_resbuf, &tmp, sizeof(tmp));\n" .
		"\ttmp = sizeof(*$arg_name[$arg]) * " .
		($arg_type[$arg_size[$arg]] eq "ptr" ? "*" : "") .
		"$arg_name[$arg_size[$arg]];\n" .
		"\tmemcpy(current_resbuf + sizeof(tmp), &tmp, sizeof(tmp));\n".
		"\t*total_len = __zoid_0c_buf_align(*total_len);\n" .
		"\t*total_len += tmp;\n";
	}
	if ($arg_type[$arg] eq "arr2d")
	{
	    print SERVER_OUT
		"\ttmp = " .
		($arg_type[$arg_size[$arg_size[$arg]]] eq "ptr" ? "*" : "") .
		"$arg_name[$arg_size[$arg_size[$arg]]];\n" .
		"\tmemcpy(current_resbuf, &tmp, sizeof(tmp));\n" .
		"\ttmp = sizeof(**$arg_name[$arg]) * " .
		"$arg_name[$arg_size[$arg]]"."[0];\n " .
		"\tmemcpy(current_resbuf + sizeof(tmp), &tmp, sizeof(tmp));\n".
		"\t*total_len = __zoid_0c_buf_align(*total_len);\n" .
		"\t*total_len += tmp;\n" .
		"\t*total_len += __zoid_0c_out_buf_ftr(0, " .
		($arg_type[$arg_size[$arg_size[$arg]]] eq "ptr" ? "*" : "") .
		"$arg_name[$arg_size[$arg_size[$arg]]]);\n" .
		"\tfor (i = 0; i < " .
		($arg_type[$arg_size[$arg_size[$arg]]] eq "ptr" ? "*" : "") .
		"$arg_name[$arg_size[$arg_size[$arg]]] - 1; i++)\n" .
		"\t{\n" .
		"\t    tmp = sizeof(**$arg_name[$arg]) * " .
		"$arg_name[$arg_size[$arg]]"."[i + 1];\n" .
		"\t    memcpy($arg_name[$arg]"."[i] + " .
		"$arg_name[$arg_size[$arg]]"."[i], &tmp, sizeof(tmp));\n" .
		"\t    *total_len = __zoid_0c_buf_align(*total_len);\n" .
		"\t    *total_len += tmp;\n" .
		"\t    *total_len += __zoid_0c_out_buf_ftr(i + 1, " .
		($arg_type[$arg_size[$arg_size[$arg]]] eq "ptr" ? "*" : "") .
		"$arg_name[$arg_size[$arg_size[$arg]]]);\n" .
		"\t}\n";
	}
	print SERVER_OUT
	    "    }\n";
    }
    else
    {
	# Data of output arrays is already in place, but we still need
	# to calculate the offsets, since non-0c arrays can be followed by
	# other data.
	if ($arg_type[$arg] eq "arr")
	{
	    print SERVER_OUT "\n" .
#		"    memcpy(current_resbuf, $arg_name[$arg], " .
#		"sizeof(*$arg_name[$arg]) * " .
#		($arg_type[$arg_size[$arg]] eq "ptr" ? "*" : "") .
#		"$arg_name[$arg_size[$arg]]);\n" .
		"    current_resbuf += sizeof(*$arg_name[$arg]) * " .
		($arg_type[$arg_size[$arg]] eq "ptr" ? "*" : "") .
		"$arg_name[$arg_size[$arg]];\n";
	}
	if ($arg_type[$arg] eq "arr2d")
	{
	    print SERVER_OUT "\n" .
		"    {\n" .
		"\tint i;\n" .
		"\tfor (i = 0; i < " .
		($arg_type[$arg_size[$arg_size[$arg]]] eq "ptr" ? "*" : "") .
		"$arg_name[$arg_size[$arg_size[$arg]]]; i++)\n" .
#		"\t{\n" .
#		"\t    memcpy(current_resbuf, $arg_name[$arg]"."[i], " .
#		"sizeof(**$arg_name[$arg]) * " .
#		"$arg_name[$arg_size[$arg]]"."[i]);\n" .
		"\t    current_resbuf += sizeof(**$arg_name[$arg]) * " .
		"$arg_name[$arg_size[$arg]]"."[i];\n" .
#		"\t}\n" .
		"    }\n";
	}
    }

    print SERVER_OUT "    }\n" if ($arg_nullok[$arg]);

    $arg_copied[$arg] = 1;
}

# Used in server stubs, to prepare a result buffer before invoking the call.
# Essentially, this copies inout arrays to the result buffer, since there's
# no other way to handle them.
sub prepare_result_buf
{
    my $arg = $_[0];

    return if ($arg_type[$arg] !~ /^arr/ || $arg_inout[$arg] !~ /out$/);

    print SERVER_OUT "\n";
    if ($arg_nullok[$arg])
    {
	print SERVER_OUT "    if ($arg_name[$arg])\n" .
	    "    {\n";
    }

    if ($arg_type[$arg] eq "arr")
    {
	if ($arg_inout[$arg] =~ /^in/)
	{
	    # Copy the array over to the output buffer.
	    print SERVER_OUT
		"    memcpy(result_buffer + $arg_name[$arg]_offset, " .
		"$arg_name[$arg], sizeof(*$arg_name[$arg]) * " .
		($arg_type[$arg_size[$arg]] eq "ptr" ? "*" : "") .
		"$arg_name[$arg_size[$arg]]);\n";
	}
	if ($arg != $userbuf_out)
	{
	    print SERVER_OUT
		"    $arg_name[$arg] = (void*)result_buffer + " .
		"$arg_name[$arg]_offset;\n";
	}
	else
	{
	    print SERVER_OUT
		"    $arg_name[$arg] = NULL;\n";
	}
    }

    if ($arg_type[$arg] eq "arr2d")
    {
	if ($arg_inout[$arg] !~ /^in/)
	{
	    # If this is output-only, then we didn't allocate the major
	    # dimension array yet, so we have to do it now.
	    print SERVER_OUT
		"    $arg_name[$arg] = alloca(sizeof(*$arg_name[$arg]) * " .
		($arg_type[$arg_size[$arg_size[$arg]]] eq "ptr" ? "*" : "") .
		"$arg_name[$arg_size[$arg_size[$arg]]]);\n";
	}

	print SERVER_OUT
	    "    {\n" .
	    "\tint i;\n" .
	    "\tfor (i = 0; i < " .
	    ($arg_type[$arg_size[$arg_size[$arg]]] eq "ptr" ? "*" : "") .
	    "$arg_name[$arg_size[$arg_size[$arg]]]; i++)\n" .
	    "\t{\n";
	if ($arg == $zerocopy_out)
	{
	    print SERVER_OUT
		"\t    $arg_name[$arg]_offset = " .
		"__zoid_0c_buf_align($arg_name[$arg]_offset);\n";
	}
	if ($arg_inout[$arg] =~ /^in/)
	{
	    # Copy the array over to the output buffer.
	    print SERVER_OUT
		"\t    memcpy(result_buffer + $arg_name[$arg]_offset, " .
		"$arg_name[$arg]"."[i], sizeof(**$arg_name[$arg]) * " .
		($arg_type[$arg_size[$arg_size[$arg]]] eq "ptr" ? "*" : "") .
		"$arg_name[$arg_size[$arg_size[$arg]]]);\n";
	}
	print SERVER_OUT
	    "\t    $arg_name[$arg]"."[i] = result_buffer + " .
	    "$arg_name[$arg]_offset;\n" .
	    "\t    $arg_name[$arg]_offset += sizeof(**$arg_name[$arg]) * " .
	    "$arg_name[$arg_size[$arg]]"."[i];\n";
	if ($arg == $zerocopy_out)
	{
	    print SERVER_OUT
		"\t    $arg_name[$arg]_offset += __zoid_0c_out_buf_ftr(i, " .
		"$arg_name[$arg_size[$arg_size[$arg]]]);\n";
	}
	print SERVER_OUT
	    "\t}\n" .
	    "    }\n";
    }

    print SERVER_OUT "    }\n" if ($arg_nullok[$arg]);
}
