/****************************************************************************/
/* ZEPTOOS:zepto-info */
/*     This file is part of ZeptoOS: The Small Linux for Big Computers.
 *     See www.mcs.anl.gov/zeptoos for more information.
 */
/* ZEPTOOS:zepto-info */
/* */
/* ZEPTOOS:zepto-fillin */
/*     $Id: dir_file.c,v 1.8 2007/06/27 15:10:45 iskra Exp $
 *     ZeptoOS_Version: 1.2
 *     ZeptoOS_Heredity: FOSS_ORIG
 *     ZeptoOS_License: GPL
 */
/* ZEPTOOS:zepto-fillin */
/* */
/* ZEPTOOS:zepto-gpl */
/*      Copyright: Argonne National Laboratory, Department of Energy,
 *                 and UChicago Argonne, LLC.  2004, 2005, 2006, 2007
 *      ZeptoOS License: GPL
 * 
 *      This software is free.  See the file ZeptoOS/misc/license.GPL
 *      for complete details on your rights to copy, modify, and use this
 *      software.
 */
/* ZEPTOOS:zepto-gpl */
/****************************************************************************/

#include "unix.h"

#include <string.h>

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/sysmacros.h>
#include <unistd.h>

#include "unix_internal.h"

int unix_access(const char *pathname /* in:str */,
		int mode /* in:obj */) /* shortout */
{
    int ret;
    int cwd;

    unix__cwd_mask_lock();

    if ((cwd = unix__getcwd_fd()) != -1)
	fchdir(cwd);

    if ((ret = access(pathname, mode)) == -1)
	ret = -errno;

    unix__cwd_mask_unlock();

    return ret;
}

int unix_chmod(const char *path /* in:str */,
	       mode_t mode /* in:obj */) /* shortout */
{
    int ret;
    int cwd;

    unix__cwd_mask_lock();

    if ((cwd = unix__getcwd_fd()) != -1)
	fchdir(cwd);

    if ((ret = chmod(path, mode)) == -1)
	ret = -errno;

    unix__cwd_mask_unlock();

    return ret;
}

int unix_fchmod(int fildes /* in:obj */,
		mode_t mode /* in:obj */) /* short */
{
    if ((fildes = unix__translate_client_fd(fildes)) == -1)
	return -errno;

    if (fchmod(fildes, mode) == -1)
	return -errno;
    else
	return 0;
}

int unix_stat(const char *path /* in:str */,
	      mode_t *st_mode /* out:ptr */,
	      ino_t *st_ino /* out:ptr */,
	      dev_t *st_dev /* out:ptr */,
	      nlink_t *st_nlink /* out:ptr */,
	      uid_t *st_uid /* out:ptr */,
	      gid_t *st_gid /* out:ptr */,
	      off_t *st_size /* out:ptr */,
	      time_t *stat_atime /* out:ptr */,
	      time_t *stat_mtime /* out:ptr */,
	      time_t *stat_ctime /* out:ptr */,
	      dev_t *st_rdev /* out:ptr */,
	      __blksize_t *st_blksize /* out:ptr */,
	      blkcnt_t *st_blocks /* out:ptr */) /* shortout */
{
    int ret;
    int cwd;
    struct stat buf;

    unix__cwd_mask_lock();

    if ((cwd = unix__getcwd_fd()) != -1)
	fchdir(cwd);

    if ((ret = stat(path, &buf)) == -1)
	ret = -errno;
    else
    {
	*st_mode = buf.st_mode;
	*st_ino = buf.st_ino;
	*st_dev = buf.st_dev;
	*st_nlink = buf.st_nlink;
	*st_uid = buf.st_uid;
	*st_gid = buf.st_gid;
	*st_size = buf.st_size;
	*stat_atime = buf.st_atime;
	*stat_mtime = buf.st_mtime;
	*stat_ctime = buf.st_ctime;
	*st_rdev = buf.st_rdev;
	*st_blksize = buf.st_blksize;
	*st_blocks = buf.st_blocks;
    }

    unix__cwd_mask_unlock();

    return ret;
}

int unix_fstat(int filedes /* in:obj */,
	       mode_t *st_mode /* out:ptr */,
	       ino_t *st_ino /* out:ptr */,
	       dev_t *st_dev /* out:ptr */,
	       nlink_t *st_nlink /* out:ptr */,
	       uid_t *st_uid /* out:ptr */,
	       gid_t *st_gid /* out:ptr */,
	       off_t *st_size /* out:ptr */,
	       time_t *stat_atime /* out:ptr */,
	       time_t *stat_mtime /* out:ptr */,
	       time_t *stat_ctime /* out:ptr */,
	       dev_t *st_rdev /* out:ptr */,
	       __blksize_t *st_blksize /* out:ptr */,
	       blkcnt_t *st_blocks /* out:ptr */) /* short */
{
    struct stat buf;

    if ((filedes = unix__translate_client_fd(filedes)) == -1)
	return -errno;

    if (filedes == FD_DEFAULT_STDOUT || filedes == FD_DEFAULT_STDERR)
    {
	if (stat("/dev/ptmx", &buf) == -1)
	    return -errno;
	/* Ensure line buffering (see glibc/sysdeps/blrts/isatty.c).  */
	*st_rdev = makedev(2, 0);
    }
    else
    {
	if (fstat(filedes, &buf) == -1)
	    return -errno;
	*st_rdev = buf.st_rdev;
    }

    *st_mode = buf.st_mode;
    *st_ino = buf.st_ino;
    *st_dev = buf.st_dev;
    *st_nlink = buf.st_nlink;
    *st_uid = buf.st_uid;
    *st_gid = buf.st_gid;
    *st_size = buf.st_size;
    *stat_atime = buf.st_atime;
    *stat_mtime = buf.st_mtime;
    *stat_ctime = buf.st_ctime;
    *st_blksize = buf.st_blksize;
    *st_blocks = buf.st_blocks;

    return 0;
}

int unix_lstat(const char *path /* in:str */,
	       mode_t *st_mode /* out:ptr */,
	       ino_t *st_ino /* out:ptr */,
	       dev_t *st_dev /* out:ptr */,
	       nlink_t *st_nlink /* out:ptr */,
	       uid_t *st_uid /* out:ptr */,
	       gid_t *st_gid /* out:ptr */,
	       off_t *st_size /* out:ptr */,
	       time_t *stat_atime /* out:ptr */,
	       time_t *stat_mtime /* out:ptr */,
	       time_t *stat_ctime /* out:ptr */,
	       dev_t *st_rdev /* out:ptr */,
	       __blksize_t *st_blksize /* out:ptr */,
	       blkcnt_t *st_blocks /* out:ptr */) /* shortout */
{
    struct stat buf;
    int ret;
    int cwd;

    unix__cwd_mask_lock();

    if ((cwd = unix__getcwd_fd()) != -1)
	fchdir(cwd);

    if ((ret = lstat(path, &buf)) == -1)
	ret = -errno;
    else
    {
	*st_mode = buf.st_mode;
	*st_ino = buf.st_ino;
	*st_dev = buf.st_dev;
	*st_nlink = buf.st_nlink;
	*st_uid = buf.st_uid;
	*st_gid = buf.st_gid;
	*st_size = buf.st_size;
	*stat_atime = buf.st_atime;
	*stat_mtime = buf.st_mtime;
	*stat_ctime = buf.st_ctime;
	*st_rdev = buf.st_rdev;
	*st_blksize = buf.st_blksize;
	*st_blocks = buf.st_blocks;
    }

    unix__cwd_mask_unlock();

    return ret;
}

int unix_stat64(const char *path /* in:str */,
		mode_t *st_mode /* out:ptr */,
		ino64_t *st_ino /* out:ptr */,
		dev_t *st_dev /* out:ptr */,
		nlink_t *st_nlink /* out:ptr */,
		uid_t *st_uid /* out:ptr */,
		gid_t *st_gid /* out:ptr */,
		off64_t *st_size /* out:ptr */,
		time_t *stat_atime /* out:ptr */,
		time_t *stat_mtime /* out:ptr */,
		time_t *stat_ctime /* out:ptr */,
		dev_t *st_rdev /* out:ptr */,
		__blksize_t *st_blksize /* out:ptr */,
		blkcnt64_t *st_blocks /* out:ptr */) /* shortout */
{
    struct stat64 buf;
    int ret;
    int cwd;

    unix__cwd_mask_lock();

    if ((cwd = unix__getcwd_fd()) != -1)
	fchdir(cwd);

    if ((ret = stat64(path, &buf)) == -1)
	ret = -errno;
    else
    {
	*st_mode = buf.st_mode;
	*st_ino = buf.st_ino;
	*st_dev = buf.st_dev;
	*st_nlink = buf.st_nlink;
	*st_uid = buf.st_uid;
	*st_gid = buf.st_gid;
	*st_size = buf.st_size;
	*stat_atime = buf.st_atime;
	*stat_mtime = buf.st_mtime;
	*stat_ctime = buf.st_ctime;
	*st_rdev = buf.st_rdev;
	*st_blksize = buf.st_blksize;
	*st_blocks = buf.st_blocks;
    }

    unix__cwd_mask_unlock();

    return ret;
}

int unix_fstat64(int filedes /* in:obj */,
		 mode_t *st_mode /* out:ptr */,
		 ino64_t *st_ino /* out:ptr */,
		 dev_t *st_dev /* out:ptr */,
		 nlink_t *st_nlink /* out:ptr */,
		 uid_t *st_uid /* out:ptr */,
		 gid_t *st_gid /* out:ptr */,
		 off64_t *st_size /* out:ptr */,
		 time_t *stat_atime /* out:ptr */,
		 time_t *stat_mtime /* out:ptr */,
		 time_t *stat_ctime /* out:ptr */,
		 dev_t *st_rdev /* out:ptr */,
		 __blksize_t *st_blksize /* out:ptr */,
		 blkcnt64_t *st_blocks /* out:ptr */) /* short */
{
    struct stat64 buf;

    if ((filedes = unix__translate_client_fd(filedes)) == -1)
	return -errno;

    if (filedes == FD_DEFAULT_STDOUT || filedes == FD_DEFAULT_STDERR)
    {
	if (stat64("/dev/ptmx", &buf) == -1)
	    return -errno;
	/* Ensure line buffering (see glibc/sysdeps/blrts/isatty.c).  */
	*st_rdev = makedev(2, 0);
    }
    else
    {
	if (fstat64(filedes, &buf) == -1)
	    return -errno;
	*st_rdev = buf.st_rdev;
    }

    *st_mode = buf.st_mode;
    *st_ino = buf.st_ino;
    *st_dev = buf.st_dev;
    *st_nlink = buf.st_nlink;
    *st_uid = buf.st_uid;
    *st_gid = buf.st_gid;
    *st_size = buf.st_size;
    *stat_atime = buf.st_atime;
    *stat_mtime = buf.st_mtime;
    *stat_ctime = buf.st_ctime;
    *st_blksize = buf.st_blksize;
    *st_blocks = buf.st_blocks;

    return 0;
}

int unix_lstat64(const char *path /* in:str */,
		 mode_t *st_mode /* out:ptr */,
		 ino64_t *st_ino /* out:ptr */,
		 dev_t *st_dev /* out:ptr */,
		 nlink_t *st_nlink /* out:ptr */,
		 uid_t *st_uid /* out:ptr */,
		 gid_t *st_gid /* out:ptr */,
		 off64_t *st_size /* out:ptr */,
		 time_t *stat_atime /* out:ptr */,
		 time_t *stat_mtime /* out:ptr */,
		 time_t *stat_ctime /* out:ptr */,
		 dev_t *st_rdev /* out:ptr */,
		 __blksize_t *st_blksize /* out:ptr */,
		 blkcnt64_t *st_blocks /* out:ptr */) /* shortout */
{
    struct stat64 buf;
    int ret;
    int cwd;

    unix__cwd_mask_lock();

    if ((cwd = unix__getcwd_fd()) != -1)
	fchdir(cwd);

    if ((ret = lstat64(path, &buf)) == -1)
	ret = -errno;
    else
    {
	*st_mode = buf.st_mode;
	*st_ino = buf.st_ino;
	*st_dev = buf.st_dev;
	*st_nlink = buf.st_nlink;
	*st_uid = buf.st_uid;
	*st_gid = buf.st_gid;
	*st_size = buf.st_size;
	*stat_atime = buf.st_atime;
	*stat_mtime = buf.st_mtime;
	*stat_ctime = buf.st_ctime;
	*st_rdev = buf.st_rdev;
	*st_blksize = buf.st_blksize;
	*st_blocks = buf.st_blocks;
    }

    unix__cwd_mask_unlock();

    return ret;
}

int unix_utime(const char *filename /* in:str */,
	       const struct utimbuf *buf /* in:ptr */) /* shortout */
{
    int ret;
    int cwd;

    unix__cwd_mask_lock();

    if ((cwd = unix__getcwd_fd()) != -1)
	fchdir(cwd);

    if ((ret = utime(filename, buf)) == -1)
	ret = -errno;

    unix__cwd_mask_unlock();

    return ret;
}

int unix_chdir(const char *path /* in:str */) /* shortout */
{
    int ret;
    int cwd;

    unix__cwd_mask_lock();

    if ((cwd = unix__getcwd_fd()) != -1)
	fchdir(cwd);

    if ((ret = chdir(path)) == -1)
	ret = -errno;

    if (ret >= 0)
    {
	cwd = open(".", O_RDONLY);

	if (cwd == -1)
	{
	    ret = -errno;

	    unix__cwd_mask_unlock();

	    return ret;
	}

	unix__setcwd_fd(cwd);
    }

    unix__cwd_mask_unlock();

    return ret;
}

int unix_fchdir(int fd /* in:obj */) /* short */
{
    int ret;

    if ((fd = unix__translate_client_fd(fd)) == -1)
	return -errno;

    if ((ret = fchdir(fd)) == -1)
	ret = -errno;

    if (ret >= 0)
    {
	int cwd;

	if ((cwd = dup(fd)) == -1)
	{
	    ret = -errno;

	    return ret;
	}

	unix__setcwd_fd(cwd);
    }

    return ret;
}

int unix_mkdir(const char *pathname /* in:str */,
	       mode_t mode /* in:obj */) /* shortout */
{
    int ret;
    int cwd;

    unix__cwd_mask_lock();

    if ((cwd = unix__getcwd_fd()) != -1)
	fchdir(cwd);

    umask(unix__getumask());

    if ((ret = mkdir(pathname, mode)) == -1)
	ret = -errno;

    unix__cwd_mask_unlock();

    return ret;
}

int unix_rmdir(const char *pathname /* in:str */) /* shortout */
{
    int ret;
    int cwd;

    unix__cwd_mask_lock();

    if ((cwd = unix__getcwd_fd()) != -1)
	fchdir(cwd);

    if ((ret = rmdir(pathname)) == -1)
	ret = -errno;

    unix__cwd_mask_unlock();

    return ret;
}

int unix_getcwd(char *buf /* out:arr:size=+1 */,
		size_t size /* in:obj */) /* shortin */
{
    int ret;
    int cwd;

    unix__cwd_mask_lock();

    if ((cwd = unix__getcwd_fd()) == -1)
	return -ENOENT;

    fchdir(cwd);

    if (!getcwd(buf, size))
	ret = -errno;
    else
	ret = strlen(buf);

    unix__cwd_mask_unlock();

    return ret;
}

int unix_chown(const char *path /* in:str */,
	       uid_t owner /* in:obj */,
	       gid_t group /* in:obj */) /* short */
{
    int ret;
    int cwd;

    unix__cwd_mask_lock();

    if ((cwd = unix__getcwd_fd()) != -1)
	fchdir(cwd);

    if ((ret = chown(path, owner, group)) == -1)
	ret = -errno;

    unix__cwd_mask_unlock();

    return ret;
}

int unix_fchown(int fd /* in:obj */,
		uid_t owner /* in:obj */,
		gid_t group /* in:obj */) /* short */
{
    if ((fd = unix__translate_client_fd(fd)) == -1)
	return -errno;

    if (fchown(fd, owner, group) == -1)
	return -errno;
    else
	return 0;
}

int unix_lchown(const char *path /* in:str */,
		uid_t owner /* in:obj */,
		gid_t group /* in:obj */) /* short */
{
    int ret;
    int cwd;

    unix__cwd_mask_lock();

    if ((cwd = unix__getcwd_fd()) != -1)
	fchdir(cwd);

    if ((ret = lchown(path, owner, group)) == -1)
	ret = -errno;

    unix__cwd_mask_unlock();

    return ret;
}

ssize_t unix_getdents(int fd /* in:obj */,
		      char *buf /* out:arr:size=+1 */,
		      size_t nbytes /* in:obj */) /* shortin */
{
    ssize_t ret;
    off_t ignore;

    if ((fd = unix__translate_client_fd(fd)) == -1)
	return -errno;

    /* We can't call getdents because libc does not export it.  However,
       it does export a similar getdirentries, which, if you look at its
       source in libc, is actually a small wrapper over getdents, so we
       should be fine.  */
    if ((ret = getdirentries(fd, buf, nbytes, &ignore)) == -1)
	return -errno;
    else
	return ret;
}

ssize_t unix_getdents64(int fd /* in:obj */,
			char *buf /* out:arr:size=+1 */,
			size_t nbytes /* in:obj */) /* shortin */
{
    ssize_t ret;
    off64_t ignore;

    if ((fd = unix__translate_client_fd(fd)) == -1)
	return -errno;

    if ((ret = getdirentries64(fd, buf, nbytes, &ignore)) == -1)
	return -errno;
    else
	return ret;
}
