/****************************************************************************/
/* ZEPTOOS:zepto-info */
/*     This file is part of ZeptoOS: The Small Linux for Big Computers.
 *     See www.mcs.anl.gov/zeptoos for more information.
 */
/* ZEPTOOS:zepto-info */
/* */
/* ZEPTOOS:zepto-fillin */
/*     $Id: unix_internal.c,v 1.7 2007/06/27 15:10:45 iskra Exp $
 *     ZeptoOS_Version: 1.2
 *     ZeptoOS_Heredity: FOSS_ORIG
 *     ZeptoOS_License: GPL
 */
/* ZEPTOOS:zepto-fillin */
/* */
/* ZEPTOOS:zepto-gpl */
/*      Copyright: Argonne National Laboratory, Department of Energy,
 *                 and UChicago Argonne, LLC.  2004, 2005, 2006, 2007
 *      ZeptoOS License: GPL
 * 
 *      This software is free.  See the file ZeptoOS/misc/license.GPL
 *      for complete details on your rights to copy, modify, and use this
 *      software.
 */
/* ZEPTOOS:zepto-gpl */
/****************************************************************************/

#include <assert.h>
#include <stdlib.h>

#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <unistd.h>

#include "unix_internal.h"

#include "zoid_api.h"

/* Descriptor 0 is reserved, since we don't support it.  */
#define FIRST_VALID_FD 1

struct client_entry
{
    int *client_fds;
    int client_fds_len;
    int cwd_fd;
    mode_t mask;
    pthread_mutex_t mutex;
};

/* Just for efficiency, a minimum client_fds len.  */
#define MIN_FDS_LEN 20

static struct client_entry* clients = NULL;
static int nclients;

static pthread_mutex_t cwd_mask_mutex = PTHREAD_MUTEX_INITIALIZER;

void unix__init(int pset_proc_count)
{
    int i;

    if (pset_proc_count < 1)
	return;

    nclients = pset_proc_count;
    clients = malloc(sizeof(struct client_entry) * nclients);
    assert(clients);

    for (i = 0; i < nclients; i++)
    {
	int j;
	clients[i].client_fds = malloc(MIN_FDS_LEN *
				       sizeof(clients[i].client_fds));
	assert(clients[i].client_fds);
	clients[i].client_fds_len = MIN_FDS_LEN;
	clients[i].client_fds[0] = 0;
	clients[i].client_fds[1] = FD_DEFAULT_STDOUT;
	clients[i].client_fds[2] = FD_DEFAULT_STDERR;
	for (j = 3; j < MIN_FDS_LEN; j++)
	    clients[i].client_fds[j] = FD_UNUSED;
	clients[i].cwd_fd = open(".", O_RDONLY);
	clients[i].mask = 022;
	pthread_mutex_init(&clients[i].mutex, NULL);
    }
}

void unix__fini(void)
{
    if (clients)
    {
	int i;

	for (i = 0; i < nclients; i++)
	{
	    if (clients[i].client_fds)
	    {
		int j;

		for (j = FIRST_VALID_FD; j < clients[i].client_fds_len; j++)
		    if (clients[i].client_fds[j] >= 0)
			close(clients[i].client_fds[j]);

		free(clients[i].client_fds);
	    }

	    if (clients[i].cwd_fd != -1)
		close(clients[i].cwd_fd);

	    pthread_mutex_destroy(&clients[i].mutex);
	}

	free(clients);
	clients = NULL;
    }
}

/*
 * Records a new file descriptor.  Translates server to client fd.
 */
int unix__record_fd(int server_fd)
{
    struct client_entry* client;
    int i;

    client = &clients[__zoid_calling_process_id()];

    pthread_mutex_lock(&client->mutex);

    for (i = FIRST_VALID_FD; i < client->client_fds_len; i++)
	if (client->client_fds[i] == FD_UNUSED)
	    break;

    if (i >= client->client_fds_len)
    {
	/* Out of space.  Enlarge the array.  */
	int newlen, j;
	int *newptr;

	newlen = (client->client_fds_len * 3) / 2;

	if (!(newptr = realloc(client->client_fds,
			       newlen * sizeof(*client->client_fds))))
	{
	    if (server_fd >= 0)
		close(server_fd);
	    pthread_mutex_unlock(&client->mutex);
	    errno = ENOMEM;
	    return -1;
	}

	client->client_fds = newptr;

	for (j = client->client_fds_len; j < newlen; j++)
	    client->client_fds[j] = FD_UNUSED;

	client->client_fds_len = newlen;
    }

    client->client_fds[i] = server_fd;

    pthread_mutex_unlock(&client->mutex);

    return i;
}

/*
 * Records a new file descriptor at a specified client fd.
 */
int unix__record_fd_at(int server_fd, int client_fd)
{
    struct client_entry* client;

    client = &clients[__zoid_calling_process_id()];

    pthread_mutex_lock(&client->mutex);

    if (client_fd < FIRST_VALID_FD)
    {
	if (server_fd >= 0)
	    close(server_fd);
	pthread_mutex_unlock(&client->mutex);
	errno = EBADF;
	return -1;
    }

    if (client_fd >= client->client_fds_len)
    {
	/* Out of space.  Enlarge the array.  */
	int newlen, j;
	int *newptr;

	newlen = (client->client_fds_len * 3) / 2;

	if (newlen <= client_fd)
	    newlen = client_fd + 1;

	if (!(newptr = realloc(client->client_fds,
			       newlen * sizeof(*client->client_fds))))
	{
	    if (server_fd >= 0)
		close(server_fd);
	    pthread_mutex_unlock(&client->mutex);
	    errno = ENOMEM;
	    return -1;
	}

	client->client_fds = newptr;

	for (j = client->client_fds_len; j < newlen; j++)
	    client->client_fds[j] = FD_UNUSED;

	client->client_fds_len = newlen;
    }

    client->client_fds[client_fd] = server_fd;

    pthread_mutex_unlock(&client->mutex);

    return client_fd;
}

int unix__translate_client_fd(int client_fd)
{
    struct client_entry* client;

    client = &clients[__zoid_calling_process_id()];

    pthread_mutex_lock(&client->mutex);

    if (client_fd < FIRST_VALID_FD || client_fd >= client->client_fds_len ||
	client->client_fds[client_fd] == FD_UNUSED)
    {
	pthread_mutex_unlock(&client->mutex);
	errno = EBADF;
	return -1;
    }

    pthread_mutex_unlock(&client->mutex);

    return client->client_fds[client_fd];
}

void unix__clear_client_fd(int client_fd)
{
    struct client_entry* client;

    client = &clients[__zoid_calling_process_id()];

    pthread_mutex_lock(&client->mutex);

    if (client_fd >= FIRST_VALID_FD && client_fd < client->client_fds_len)
	client->client_fds[client_fd] = FD_UNUSED;

    pthread_mutex_unlock(&client->mutex);
}

int unix__getcwd_fd(void)
{
    return clients[__zoid_calling_process_id()].cwd_fd;
}

void unix__setcwd_fd(int cwd_fd)
{
    struct client_entry* client;

    client = &clients[__zoid_calling_process_id()];

    if (client->cwd_fd != -1)
	close(client->cwd_fd);

    client->cwd_fd = cwd_fd;
}

mode_t unix__getumask(void)
{
    return clients[__zoid_calling_process_id()].mask;
}

mode_t unix__setumask(mode_t mask)
{
    struct client_entry* client;
    mode_t oldmask;

    client = &clients[__zoid_calling_process_id()];

    oldmask = client->mask;

    client->mask = mask & 0777;

    return oldmask;
}

/* cwd_mask_mutex guards critical sections that in a non-atomic way
   change and depend on the current working directory (and sometimes
   also umask) of a UNIX process.  We can't let two or more threads run
   such code sections simultaneously, hence the mutex.  */
void unix__cwd_mask_lock(void)
{
    pthread_mutex_lock(&cwd_mask_mutex);
}

void unix__cwd_mask_unlock(void)
{
    pthread_mutex_unlock(&cwd_mask_mutex);
}
