
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include "zoidfs.h"
#include "zint-handler.h"

#include "pvfs2-sysint.h"
#include "pvfs2-util.h"

#define zint_pvfs2_input_handle(_zfs_handle, _pvfs2_ref) \
    do { \
        _pvfs2_ref.handle = *((PVFS_handle *) \
                              (_zfs_handle->data)); \
        _pvfs2_ref.fs_id =  *((PVFS_fs_id *) \
                              ((_zfs_handle->data) + sizeof(PVFS_handle))); \
    } while(0)

#define zint_pvfs2_output_handle(_pvfs2_ref, _zfs_handle) \
    do { \
        uint8_t * _zptr = (uint8_t *)_zfs_handle->data; \
        *(PVFS_handle *) _zptr = _pvfs2_ref.handle; \
        _zptr += sizeof(PVFS_handle); \
        *(PVFS_fs_id *) _zptr = _pvfs2_ref.fs_id; \
        _zptr += sizeof(PVFS_fs_id); \
        *(zint_handle_type_t *) _zptr = ZINT_HANDLE_TYPE_PVFS2; \
    } while(0)

#define zint_pvfs2_input_time(_ztime) \
    (((_ztime).seconds) + ((uint64_t) ((_ztime).useconds * 10e-6)))

#define zint_pvfs2_output_time(_ptime, _ztime) \
    do { \
        (_ztime).seconds = _ptime; \
        (_ztime).useconds = 0; \
    } while(0)

#define zint_pvfs2_input_creds(_sattr, _creds) \
    (_creds).uid = (_sattr)->uid; (_creds).gid = (_sattr)->gid

inline static uint32_t zint_pvfs2_input_attr_mask(uint16_t zmask);

inline static uint16_t zint_pvfs2_output_attr_mask(uint32_t pmask);

inline static int zint_pvfs2_input_sattr(const zoidfs_sattr_t * zsattr,
                                         PVFS_sys_attr * resattr);

inline static PVFS_ds_type zint_pvfs2_input_attr_type(
    zoidfs_attr_type_t ztype);

inline static int zint_pvfs2_input_attr(const zoidfs_attr_t * zattr,
                                        PVFS_credentials creds,
                                        PVFS_sys_attr * resattr);

inline static int zint_pvfs2_output_attr(PVFS_sys_attr attr,
                                         zoidfs_attr_t * zattr);

inline static int zint_pvfs2_output_error(PVFS_error error);

inline static void zint_pvfs2_split_path(const char * path,
                                         char ** base,
                                         char ** component);

inline static uint32_t zint_pvfs2_input_attr_mask(uint16_t zmask)
{
    uint32_t pmask = 0;

    if(zmask & ZOIDFS_ATTR_MODE)
    {
        pmask |= PVFS_ATTR_SYS_PERM;
    }
    
    if(zmask & ZOIDFS_ATTR_UID)
    {
        pmask |= PVFS_ATTR_SYS_UID;
    }

    if(zmask & ZOIDFS_ATTR_GID)
    {
        pmask |= PVFS_ATTR_SYS_GID;
    }

    if(zmask & ZOIDFS_ATTR_SIZE)
    {
        pmask |= PVFS_ATTR_SYS_SIZE;
    }

    if(zmask & ZOIDFS_ATTR_ATIME)
    {
        pmask |= PVFS_ATTR_SYS_ATIME;
    }

    if(zmask & ZOIDFS_ATTR_CTIME)
    {
        pmask |= PVFS_ATTR_SYS_CTIME;
    }

    if(zmask & ZOIDFS_ATTR_MTIME)
    {
        pmask |= PVFS_ATTR_SYS_MTIME;
    }

    return pmask;
}

inline static uint16_t zint_pvfs2_output_attr_mask(uint32_t pmask)
{
    uint16_t zmask = 0;

    if(pmask & PVFS_ATTR_SYS_PERM)
    {
        pmask |= ZOIDFS_ATTR_MODE;
    }
    
    if(pmask & PVFS_ATTR_SYS_UID)
    {
        zmask |= ZOIDFS_ATTR_UID;
    }

    if(pmask & PVFS_ATTR_SYS_GID)
    {
        zmask |= ZOIDFS_ATTR_GID;
    }

    if(pmask & PVFS_ATTR_SYS_SIZE)
    {
        zmask |= ZOIDFS_ATTR_SIZE;
    }

    if(pmask & PVFS_ATTR_SYS_ATIME)
    {
        zmask |= ZOIDFS_ATTR_ATIME;
    }

    if(pmask & PVFS_ATTR_SYS_CTIME)
    {
        zmask |= ZOIDFS_ATTR_CTIME;
    }

    if(pmask & PVFS_ATTR_SYS_MTIME)
    {
        zmask |= ZOIDFS_ATTR_MTIME;
    }

    return zmask;
}

inline static PVFS_ds_type zint_pvfs2_input_attr_type(
    zoidfs_attr_type_t ztype)
{
    if(ztype == ZOIDFS_REG)
    {
        return PVFS_TYPE_METAFILE;
    }
    else if(ztype == ZOIDFS_DIR)
    {
        return PVFS_TYPE_DIRECTORY;
    }
    return -1;
}

inline static zoidfs_attr_type_t zint_pvfs2_output_attr_type(
    PVFS_ds_type ptype)
{
    if(ptype == PVFS_TYPE_METAFILE)
    {
        return ZOIDFS_REG;
    }
    else if(ptype == PVFS_TYPE_DIRECTORY)
    {
        return ZOIDFS_DIR;
    }
    return -1;
}

inline static int zint_pvfs2_input_sattr(const zoidfs_sattr_t * zsattr,
                                         PVFS_sys_attr * resattr)
{
    PVFS_sys_attr pattr;

    pattr.owner = zsattr->uid;
    pattr.group = zsattr->gid;
    pattr.perms = zsattr->mode;
    pattr.atime = zint_pvfs2_input_time(zsattr->atime);
    pattr.mtime = zint_pvfs2_input_time(zsattr->mtime);
    pattr.size  = zsattr->size;
    pattr.mask = zint_pvfs2_input_attr_mask(zsattr->mask);

    *resattr = pattr;

    return ZFS_OK;
}

inline static int zint_pvfs2_input_attr(const zoidfs_attr_t * zattr,
                                        PVFS_credentials creds,
                                        PVFS_sys_attr * pattr)
{
    pattr->owner = creds.uid;
    pattr->group = creds.gid;
    pattr->perms = zattr->mode;
    pattr->atime = zint_pvfs2_input_time(zattr->atime);
    pattr->mtime = zint_pvfs2_input_time(zattr->mtime);
    pattr->size  = zattr->size;

    pattr->link_target = NULL;
    pattr->dfile_count = 0;
    pattr->dirent_count = 0;
    pattr->objtype = zint_pvfs2_input_attr_type(zattr->type);
    pattr->mask = zint_pvfs2_input_attr_mask(zattr->mask);

    return ZFS_OK;
}

inline static int zint_pvfs2_output_attr(PVFS_sys_attr attr,
                                         zoidfs_attr_t * zattr)
{
    zattr->mode = attr.perms;
    zint_pvfs2_output_time(attr.atime, zattr->atime);
    zint_pvfs2_output_time(attr.mtime, zattr->mtime);
    zattr->size = attr.size;

    zattr->type = zint_pvfs2_output_attr_type(attr.objtype);
    zattr->mask = zint_pvfs2_output_attr_mask(attr.mask);

    return ZFS_OK;
}

inline static int zint_pvfs2_output_error(PVFS_error error)
{
    switch(-error)
    {
    case PVFS_EPERM:
        return ZFSERR_PERM;
    case PVFS_ENOENT:
        return ZFSERR_NOENT;
    case PVFS_EIO:
        return ZFSERR_IO;
    case PVFS_ENXIO:
        return ZFSERR_NXIO;
    case PVFS_EACCES:
        return ZFSERR_ACCES;
    case PVFS_EEXIST:
        return ZFSERR_EXIST;
    case PVFS_ENODEV:
        return ZFSERR_NODEV;
    case PVFS_ENOTDIR:
        return ZFSERR_NOTDIR;
    case PVFS_EISDIR:
        return ZFSERR_ISDIR;
    case PVFS_EFBIG:
        return ZFSERR_FBIG;
    case PVFS_ENOSPC:
        return ZFSERR_NOSPC;
    case PVFS_EROFS:
        return ZFSERR_ROFS;
    case PVFS_ENAMETOOLONG:
        return ZFSERR_NAMETOOLONG;
    case PVFS_ENOTEMPTY:
        return ZFSERR_NOTEMPTY;
    }

    return -error;
}

inline static void zint_pvfs2_split_path(const char * path,
                                         char ** base,
                                         char ** component)
{
    char * tmpstr;
    char * index;

    assert(path);

    tmpstr = strdup(path);

    index = rindex(tmpstr, '/');
    if(index)
    {
        *index = 0;
        ++index;

        *component = strdup(index);
        *base = tmpstr;
    }
    else
    {
        *component = tmpstr;
        *base = NULL;
    }
}

void zint_pvfs2_null(void)
{
}

int zint_pvfs2_create(const zoidfs_handle_t * parent_handle,
                      const char * component_name,
                      const char * full_path,
                      const zoidfs_sattr_t * sattr,
                      zoidfs_handle_t * handle,
                      int * created)
{
    int ret;
    PVFS_object_ref ref;
    PVFS_sys_attr pattr;
    PVFS_credentials creds;
    PVFS_sysresp_create create_resp;
    PVFS_sysresp_lookup lookup_resp;

    zint_pvfs2_input_creds(sattr, creds);

    ret = zint_pvfs2_input_sattr(sattr, &pattr);
    if(ret < 0)
    {
        return zint_pvfs2_output_error(ret);
    }

    if(!(pattr.mask & PVFS_ATTR_SYS_CTIME))
    {
        pattr.mask |= PVFS_ATTR_SYS_CTIME;
        pattr.ctime = pattr.mtime;
    }

    zint_pvfs2_input_handle(parent_handle, ref);

    if(component_name && parent_handle)
    {
        /* use the default distribution */
        ret = PVFS_sys_create(
            (char *)component_name, ref, pattr, &creds, NULL, &create_resp);
    }
    else if(full_path)
    {
        char * component_str;
        char * base_str;

        zint_pvfs2_split_path(full_path, &base_str, &component_str);
        
        ret = PVFS_sys_lookup(
            ref.fs_id, base_str, &creds, &lookup_resp, 
            PVFS2_LOOKUP_LINK_NO_FOLLOW);

        if(base_str)
        {
            free(base_str);
        }

        if(ret < 0)
        {
            free(component_str);
            
            return zint_pvfs2_output_error(ret);
        }

        ret = PVFS_sys_create(
            (char *)component_str, lookup_resp.ref, 
            pattr, &creds, NULL, &create_resp); 
            
        free(component_str);
    }
        
    if(ret < 0)
    {
        if(ret == -PVFS_EEXIST)
        {
            *created = 0;
            ret = ZFS_OK;
        }
        else
        {
            return zint_pvfs2_output_error(ret);
        }
    }
    else
    {
        *created = 1;
    }

    zint_pvfs2_output_handle(create_resp.ref, handle);
    
    return ZFS_OK;
}

int zint_pvfs2_getattr(const zoidfs_handle_t * handle,
                       zoidfs_attr_t * attr)
{
    int ret;
    PVFS_credentials creds;
    PVFS_object_ref ref;
    PVFS_sysresp_getattr getattr_resp;

    zint_pvfs2_input_handle(handle, ref);

    PVFS_util_gen_credentials(&creds);

    ret = PVFS_sys_getattr(ref, PVFS_ATTR_SYS_ALL, &creds, &getattr_resp);
    if(ret < 0)
    {
        return zint_pvfs2_output_error(ret);
    }
    
    zint_pvfs2_output_attr(getattr_resp.attr, attr);

    return ZFS_OK;    
}

int zint_pvfs2_setattr(const zoidfs_handle_t * handle,
                       const zoidfs_sattr_t * sattr,
                       zoidfs_attr_t * attr)
{
    int ret;
    PVFS_credentials creds;
    PVFS_object_ref ref;
    PVFS_sys_attr pattr;
    PVFS_sysresp_getattr getattr_resp;

    zint_pvfs2_input_handle(handle, ref);

    zint_pvfs2_input_creds(sattr, creds);

    zint_pvfs2_input_sattr(sattr, &pattr);

    ret = PVFS_sys_setattr(ref, pattr, &creds);
    if(ret < 0)
    {
        return zint_pvfs2_output_error(ret);
    }

    ret = PVFS_sys_getattr(ref, pattr.mask, &creds, &getattr_resp);
    if(ret < 0)
    {
        return zint_pvfs2_output_error(ret);
    }

    zint_pvfs2_output_attr(getattr_resp.attr, attr);
    return ZFS_OK;
}

int zint_pvfs2_lookup(const zoidfs_handle_t * parent_handle,
                      const char * component_name,
                      const char * full_path,
                      zoidfs_handle_t * handle)
{
    int res;
    PVFS_object_ref ref;
    PVFS_credentials creds;
    PVFS_sysresp_lookup lookup_resp;

    PVFS_util_gen_credentials(&creds);

    zint_pvfs2_input_handle(parent_handle, ref);

    if(component_name)
    {
        res = PVFS_sys_ref_lookup(ref.fs_id, (char *)component_name,
                                  ref, &creds, &lookup_resp, 
                                  PVFS2_LOOKUP_LINK_NO_FOLLOW);
    }
    else
    {
        res = PVFS_sys_lookup(ref.fs_id, (char *)full_path, 
                              &creds, &lookup_resp, 
                              PVFS2_LOOKUP_LINK_NO_FOLLOW);
    }

    if(res < 0)
    {
        return zint_pvfs2_output_error(res);
    }

    zint_pvfs2_output_handle(lookup_resp.ref, handle);
    return ZFS_OK;
}

int zint_pvfs2_readlink(const zoidfs_handle_t * handle,
                        char * buffer,
                        size_t buffer_length)
{
    return -ENOSYS;
}

int zint_pvfs2_read(const zoidfs_handle_t * handle,
                    int mem_count,
                    void * mem_starts[],
                    const size_t mem_sizes[],
                    int file_count,
                    const uint64_t file_starts[],
                    uint64_t file_sizes[])
{
    int ret;
    int i;
    PVFS_object_ref ref;
    PVFS_Request mem_req;
    PVFS_Request file_req;
    PVFS_size * displacements;
    PVFS_credentials creds;
    PVFS_sysresp_io read_resp;

    displacements = malloc(sizeof(PVFS_size) * mem_count);
    if(!displacements)
    {
        return -ENOMEM;
    }

    for(i = 0; i < mem_count; ++i)
    {
        displacements[i] = (mem_starts[i] - mem_starts[0]);
    }

    ret = PVFS_Request_indexed(mem_count, (int32_t *)mem_sizes, displacements,
                               PVFS_BYTE, &mem_req);
    if(ret < 0)
    {
        free(displacements);
        return zint_pvfs2_output_error(ret);
    }

    free(displacements);

    ret = PVFS_Request_indexed(file_count, (int32_t *)file_sizes, 
                               (PVFS_size *)file_starts,
                               PVFS_BYTE, &file_req);
    if(ret < 0)
    {
        free(displacements);
        return zint_pvfs2_output_error(ret);
    }

    free(displacements);

    zint_pvfs2_input_handle(handle, ref);

    ret = PVFS_sys_read(
        ref, file_req, 0, mem_starts[0], mem_req, &creds, &read_resp);
    if(ret < 0)
    {
        return zint_pvfs2_output_error(ret);
    }

    PVFS_Request_free(&mem_req);
    PVFS_Request_free(&file_req);

    return ZFS_OK;
}

int zint_pvfs2_write(const zoidfs_handle_t * handle,
                     int mem_count,
                     const void * mem_starts[],
                     const size_t mem_sizes[],
                     int file_count,
                     const uint64_t file_starts[],
                     uint64_t file_sizes[])
{
    return -ENOSYS;
}

int zint_pvfs2_commit(const zoidfs_handle_t * handle)
{
    return -ENOSYS;
}

int zint_pvfs2_remove(const zoidfs_handle_t * parent_handle,
                      const char * component_name,
                      const char * full_path,
                      zoidfs_cache_hint_t * parent_hint)
{
    int ret;
    PVFS_object_ref ref;
    PVFS_credentials creds;
    PVFS_sysresp_lookup lookup_resp;

    PVFS_util_gen_credentials(&creds);

    zint_pvfs2_input_handle(parent_handle, ref);

    if(component_name)
    {
        ret = PVFS_sys_remove((char *)component_name, ref, &creds);
    }
    else
    {
        char * component_str;
        char * base_str;

        zint_pvfs2_split_path(full_path, &base_str, &component_str);
        
        ret = PVFS_sys_lookup(ref.fs_id, base_str, &creds, &lookup_resp, 
                              PVFS2_LOOKUP_LINK_NO_FOLLOW);
        
        if(base_str)
        {
            free(base_str);
        }

        if(ret < 0)
        {
            free(component_str);
            return zint_pvfs2_output_error(ret);
        }

        ret = PVFS_sys_remove(component_str, lookup_resp.ref, &creds);

        free(component_str);
    }

    if(ret < 0)
    {
        return zint_pvfs2_output_error(ret);
    }

    return ZFS_OK;
}

int zint_pvfs2_rename(const zoidfs_handle_t * from_parent_handle,
                      const char * from_component_name,
                      const char * from_full_path,
                      const zoidfs_handle_t * to_parent_handle,
                      const char * to_component_name,
                      const char * to_full_path,
                      zoidfs_cache_hint_t * from_parent_hint,
                      zoidfs_cache_hint_t * to_parent_hint)
{
    return -ENOSYS;
}

int zint_pvfs2_symlink(const zoidfs_handle_t * from_parent_handle,
                       const char * from_component_name,
                       const char * from_full_path,
                       const zoidfs_handle_t * to_parent_handle,
                       const char * to_component_name,
                       const char * to_full_path,
                       const zoidfs_sattr_t * attr,
                       zoidfs_cache_hint_t * from_parent_hint,
                       zoidfs_cache_hint_t * to_parent_hint)
{
    return -ENOSYS;
}

int zint_pvfs2_mkdir(const zoidfs_handle_t * parent_handle,
                     const char * component_name,
                     const char * full_path,
                     const zoidfs_sattr_t * attr,
                     zoidfs_cache_hint_t * parent_hint)
{
    int ret;
    PVFS_object_ref ref;
    PVFS_credentials creds;
    PVFS_sys_attr pattr;
    PVFS_sysresp_lookup lookup_resp;
    PVFS_sysresp_mkdir mkdir_resp;

    zint_pvfs2_input_creds(attr, creds);

    zint_pvfs2_input_sattr(attr, &pattr);

    if(component_name)
    {
        zint_pvfs2_input_handle(parent_handle, ref);

        ret = PVFS_sys_mkdir((char *)component_name, ref, 
                             pattr, &creds, &mkdir_resp);
    }
    else
    {
        PVFS_fs_id fsid;
        char * component_str;
        char * base_str;

        zint_pvfs2_split_path(full_path, &base_str, &component_str);

        ret = PVFS_sys_lookup(fsid, base_str, &creds, &lookup_resp, 
                              PVFS2_LOOKUP_LINK_NO_FOLLOW);

        if(base_str)
        {
            free(base_str);
        }

        if(ret < 0)
        {
            return zint_pvfs2_output_error(ret);
        }

        ret = PVFS_sys_mkdir(component_str, lookup_resp.ref, 
                             pattr, &creds, &mkdir_resp);

        free(component_str);
    }

    if(ret < 0)
    {
        return zint_pvfs2_output_error(ret);
    }

    return ZFS_OK;
}

int zint_pvfs2_readdir(const zoidfs_handle_t * parent_handle,
                       zoidfs_dirent_cookie_t cookie,
                       int * entry_count,
                       zoidfs_dirent_t * entries,
                       zoidfs_cache_hint_t * parent_hint)
{
    int ret;
    int i;
    PVFS_object_ref ref;
    PVFS_credentials creds;
    PVFS_sysresp_readdir readdir_resp;

    PVFS_util_gen_credentials(&creds);
    
    zint_pvfs2_input_handle(parent_handle, ref);

    ret = PVFS_sys_readdir(ref, (PVFS_ds_position)cookie, 
                           *entry_count, &creds, &readdir_resp);
    if(ret < 0)
    {
        return zint_pvfs2_output_error(ret);
    }

    *entry_count = readdir_resp.pvfs_dirent_outcount;
    for(; i < *entry_count; ++i)
    {
        zoidfs_handle_t * entry_handle;
        PVFS_object_ref entry_ref = ref;
        strncpy(entries[i].name, readdir_resp.dirent_array[i].d_name,
                ZOIDFS_NAME_MAX);
        entries[i].name[ZOIDFS_NAME_MAX] = 0;
        
        entry_ref.handle = readdir_resp.dirent_array[i].handle;
        zint_pvfs2_output_handle(entry_ref,
                                 entry_handle);
        entries[i].handle = *entry_handle;
        entries[i].cookie = (zoidfs_dirent_cookie_t)readdir_resp.token;
    }

    free(readdir_resp.dirent_array);

    return ZFS_OK;
}

int zint_pvfs2_resize(const zoidfs_handle_t * handle,
                      uint64_t size)
{
    int ret;
    PVFS_object_ref ref;
    PVFS_credentials creds;

    PVFS_util_gen_credentials(&creds);

    zint_pvfs2_input_handle(handle, ref);

    ret = PVFS_sys_truncate(ref, size, &creds);
    if(ret < 0)
    {
        return zint_pvfs2_output_error(ret);
    }

    return ZFS_OK;
}

int zint_pvfs2_resolve_path(const char * local_path,
                            zoidfs_handle_t * handle,
                            char * fs_path,
                            int fs_path_max)
{
    int ret;
    PVFS_object_ref ref;
    PVFS_fs_id fsid;

    ret = PVFS_util_resolve(local_path, &fsid, fs_path, fs_path_max);
    if(ret < 0)
    {
        return zint_pvfs2_output_error(ret);
    }

    ref.fs_id = fsid;
    ref.handle = 0;

    if(handle)
    {
        zint_pvfs2_output_handle(ref, handle);
    }

    return ZFS_OK;
}
        

/**
 * OPTIONAL
 */
int zint_pvfs2_init(void)
{
    int ret;
    ret = PVFS_util_init_defaults();

    if(ret < 0)
    {
        return zint_pvfs2_output_error(ret);
    }

    return ZFS_OK;
}

/**
 * OPTIONAL
 */
int zint_pvfs2_finalize(void)
{
    PVFS_sys_finalize();
    return ZFS_OK;
}

zint_handler_t pvfs2_handler =
{
    ZINT_HANDLE_TYPE_PVFS2,
    zint_pvfs2_null,
    zint_pvfs2_getattr,
    zint_pvfs2_setattr,
    zint_pvfs2_lookup,
    zint_pvfs2_readlink,
    zint_pvfs2_read,
    zint_pvfs2_write,
    zint_pvfs2_commit,
    zint_pvfs2_create,
    zint_pvfs2_remove,
    zint_pvfs2_rename,
    zint_pvfs2_symlink,
    zint_pvfs2_mkdir,
    zint_pvfs2_readdir,
    zint_pvfs2_resize,
    zint_pvfs2_resolve_path,
    zint_pvfs2_init,
    zint_pvfs2_finalize
};

/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 * End:
 *
 * vim: ts=8 sts=4 sw=4 expandtab
 */
