
#include <errno.h>
#include <assert.h>
#include <string.h>
#include "zoidfs.h"
#include "zint-handler.h"

extern zint_handler_t pvfs2_handler;
extern zint_handler_t local_handler;

#define ZINT_HANDLERS_COUNT 2
zint_handler_t * zint_handlers[ZINT_HANDLERS_COUNT] = 
{
    &pvfs2_handler,
    &local_handler
};

inline static int zint_locate_handler(const zoidfs_handle_t * handle,
                                      const char * path,
                                      zint_handler_t ** handler,
                                      char * newpath,
                                      int newpath_maxlen,
                                      zoidfs_handle_t * newhandle);

inline static zint_handler_t * zint_get_handler(zint_handle_type_t type);

inline static zint_handler_t * zint_get_handler_from_path(
    const char * path,
    char * newpath,
    int newpath_maxlen,
    zoidfs_handle_t * newhandle);

void zoidfs_null(void)
{
}

int zoidfs_create(const zoidfs_handle_t * parent_handle,
		  const char * component_name,
		  const char * full_path,
		  const zoidfs_sattr_t * sattr,
		  zoidfs_handle_t * handle,
		  int * created)
{
    int ret;
    char new_path[ZOIDFS_PATH_MAX];
    zint_handler_t * handler;
    zoidfs_handle_t newhandle;

    ret = zint_locate_handler(parent_handle, full_path, &handler,
                              new_path, ZOIDFS_PATH_MAX, &newhandle);
    if(ret < 0)
    {
        return -EINVAL;
    }

    return handler->create(&newhandle,
                           component_name,
                           full_path ? new_path : NULL,
                           sattr,
                           handle,
                           created);
}

int zoidfs_getattr(const zoidfs_handle_t * handle,
                   zoidfs_attr_t * attr)
{
    int ret;
    zint_handler_t * handler;

    ret = zint_locate_handler(handle, NULL, &handler, NULL, 0, NULL);
    if(ret < 0)
    {
        return -EINVAL;
    }

    return handler->getattr(handle, attr);
}
    
int zoidfs_setattr(const zoidfs_handle_t * handle,
                   const zoidfs_sattr_t * sattr,
                   zoidfs_attr_t * attr)
{
    zint_handler_t * handler;
    int ret;

    ret = zint_locate_handler(handle, NULL, &handler, NULL, 0, NULL);
    if(ret < 0)
    {
        return -EINVAL;
    }

    return handler->setattr(handle, sattr, attr);
}
 
int zoidfs_lookup(const zoidfs_handle_t * parent_handle,
                  const char * component_name,
                  const char * full_path,
                  zoidfs_handle_t * handle)
{
    zint_handler_t * handler;
    char new_path[ZOIDFS_PATH_MAX];
    int ret;
    zoidfs_handle_t newhandle;

    ret = zint_locate_handler(parent_handle, full_path, &handler, 
                              new_path, ZOIDFS_PATH_MAX, &newhandle);
    if(ret < 0)
    {
        return -EINVAL;
    }

    return handler->lookup(&newhandle, 
                           component_name, 
                           full_path ? new_path : NULL, 
                           handle);
}

int zoidfs_readlink(const zoidfs_handle_t * handle,
                    char * buffer,
                    size_t buffer_length)
{
    zint_handler_t * handler;
    int ret;

    ret = zint_locate_handler(handle, NULL, &handler, NULL, 0, NULL);
    if(ret < 0)
    {
        return -EINVAL;
    }

    return handler->readlink(handle, buffer, buffer_length);
}

int zoidfs_read(const zoidfs_handle_t * handle,
                int mem_count,
                void * mem_starts[],
                const size_t mem_sizes[],
                int file_count,
                const uint64_t file_starts[],
                uint64_t file_sizes[])
{
    zint_handler_t * handler;
    int ret;

    ret = zint_locate_handler(handle, NULL, &handler, NULL, 0, NULL);
    if(ret < 0)
    {
        return -EINVAL;
    }

    return handler->read(handle, 
                         mem_count, 
                         mem_starts, 
                         mem_sizes, 
                         file_count, 
                         file_starts, 
                         file_sizes);
}
     
int zoidfs_write(const zoidfs_handle_t * handle,
                 int mem_count,
                 const void * mem_starts[],
                 const size_t mem_sizes[],
                 int file_count,
                 const uint64_t file_starts[],
                 uint64_t file_sizes[])
{
    zint_handler_t * handler;
    int ret;

    ret = zint_locate_handler(handle, NULL, &handler, NULL, 0, NULL);
    if(ret < 0)
    {
        return -EINVAL;
    }

    return handler->write(handle,
                          mem_count,
                          mem_starts,
                          mem_sizes,
                          file_count,
                          file_starts,
                          file_sizes);
}

int zoidfs_remove(const zoidfs_handle_t * parent_handle,
                  const char * component_name,
                  const char * full_path,
                  zoidfs_cache_hint_t * parent_hint)
{
    int ret;
    zint_handler_t * handler;
    char newpath[ZOIDFS_PATH_MAX];
    zoidfs_handle_t newhandle;
    
    ret = zint_locate_handler(parent_handle, full_path, &handler, 
                              newpath, ZOIDFS_PATH_MAX, &newhandle);
    if(ret < 0)
    {
        return -EINVAL;
    }

    return handler->remove(&newhandle,
                           component_name,
                           full_path ? newpath : NULL,
                           parent_hint);
}

int zoidfs_commit(const zoidfs_handle_t * handle)
{
    int ret;
    zint_handler_t * handler;

    ret = zint_locate_handler(handle, NULL, &handler, NULL, 0, NULL);
    if(ret < 0)
    {
        return -EINVAL;
    }

    return handler->commit(handle);
}

int zoidfs_rename(const zoidfs_handle_t * from_parent_handle,
                  const char * from_component_name,
                  const char * from_full_path,
                  const zoidfs_handle_t * to_parent_handle,
                  const char * to_component_name,
                  const char * to_full_path,
                  zoidfs_cache_hint_t * from_parent_hint,
                  zoidfs_cache_hint_t * to_parent_hint)
{
    int ret;
    zint_handler_t * handler;
    char  new_from_path[ZOIDFS_PATH_MAX], new_to_path[ZOIDFS_PATH_MAX];
    zoidfs_handle_t new_from_handle, new_to_handle;
    
    ret = zint_locate_handler(
        from_parent_handle, from_full_path, &handler, 
        new_from_path, ZOIDFS_PATH_MAX, &new_from_handle);
    if(ret < 0)
    {
        return -EINVAL;
    }

    ret = zint_locate_handler(
        to_parent_handle, to_full_path, &handler,
        new_to_path, ZOIDFS_PATH_MAX, &new_to_handle);
    if(ret < 0)
    {
        return -EINVAL;
    }

    return handler->rename(&new_from_handle,
                           from_component_name,
                           from_full_path ? new_from_path : from_full_path,
                           &new_to_handle,
                           to_component_name,
                           to_full_path ? new_to_path : to_full_path,
                           from_parent_hint,
                           to_parent_hint);
}

int zoidfs_symlink(const zoidfs_handle_t * from_parent_handle,
                   const char * from_component_name,
                   const char * from_full_path,
                   const zoidfs_handle_t * to_parent_handle,
                   const char * to_component_name,
                   const char * to_full_path,
                   const zoidfs_sattr_t * attr,
                   zoidfs_cache_hint_t * from_parent_hint,
                   zoidfs_cache_hint_t * to_parent_hint)
{
    int ret;
    zint_handler_t * handler;
    char new_from_path[ZOIDFS_PATH_MAX];
    char new_to_path[ZOIDFS_PATH_MAX];
    zoidfs_handle_t new_from_handle, new_to_handle;

    ret = zint_locate_handler(from_parent_handle, from_full_path, 
                              &handler, new_from_path, ZOIDFS_PATH_MAX,
                              &new_from_handle);
    if(ret < 0)
    {
        return -EINVAL;
    }

    ret = zint_locate_handler(to_parent_handle, from_full_path, 
                              &handler, new_to_path, ZOIDFS_PATH_MAX,
                              &new_to_handle);
    if(ret < 0)
    {
        return -EINVAL;
    }

    return handler->symlink(&new_from_handle, 
                            from_component_name,
                            (from_full_path ? new_from_path : from_full_path),
                            &new_to_handle,
                            to_component_name,
                            (to_full_path ? new_to_path : to_full_path),
                            attr,
                            from_parent_hint,
                            to_parent_hint);
}

int zoidfs_mkdir(const zoidfs_handle_t * parent_handle,
                 const char * component_name,
                 const char * full_path,
                 const zoidfs_sattr_t * attr,
                 zoidfs_cache_hint_t * parent_hint)
{
    int ret;
    zint_handler_t * handler;
    char new_path[ZOIDFS_PATH_MAX];
    zoidfs_handle_t new_handle;

    ret = zint_locate_handler(parent_handle, full_path, &handler, 
                              new_path, ZOIDFS_PATH_MAX,
                              &new_handle);
    if(ret < 0)
    {
        return -EINVAL;
    }

    return handler->mkdir(&new_handle,
                          component_name,
                          full_path ? new_path : NULL,
                          attr,
                          parent_hint);
}

int zoidfs_readdir(const zoidfs_handle_t * parent_handle,
                   zoidfs_dirent_cookie_t cookie,
                   int * entry_count,
                   zoidfs_dirent_t * entries,
                   zoidfs_cache_hint_t * parent_hint)
{
    int ret;
    zint_handler_t * handler;

    ret = zint_locate_handler(parent_handle, NULL, &handler, NULL, 0, NULL);
    if(ret < 0)
    {
        return -EINVAL;
    }

    return handler->readdir(parent_handle,
                            cookie,
                            entry_count,
                            entries,
                            parent_hint);
}

int zoidfs_resize(const zoidfs_handle_t * handle,
                  uint64_t size)
{
    int ret;
    zint_handler_t * handler;

    ret = zint_locate_handler(handle, NULL, &handler, NULL, 0, NULL);
    if(ret < 0)
    {
        return -EINVAL;
    }

    return handler->resize(handle,
                           size);
}    

/**
 * OPTIONAL
 */
int zoidfs_init(void)
{
    int ret;
    int i = 0;
    for(; i < ZINT_HANDLERS_COUNT; ++i)
    {
        ret = zint_handlers[i]->init();
        if(ret != ZFS_OK)
        {
            return ret;
        }
    }
    return ZFS_OK;
}

/**
 * OPTIONAL
 */
int zoidfs_finalize(void)
{
    int ret;
    int i = 0;

    for(; i < ZINT_HANDLERS_COUNT; ++i)
    {
        ret = zint_handlers[i]->finalize();
        if(ret != ZFS_OK)
        {
            return ret;
        }
    }
    return ZFS_OK;
}


inline static zint_handler_t * zint_get_handler(zint_handle_type_t type)
{
    assert(type < ZINT_HANDLERS_COUNT);
    return zint_handlers[(uint32_t)(type - 1)];
}

inline static zint_handler_t * zint_get_handler_from_path(
    const char * path,
    char * newpath,
    int newpath_maxlen,
    zoidfs_handle_t * newhandle)
{
    int ret;
    int i = 0;
  
    for(; i < ZINT_HANDLERS_COUNT; ++i)
    {
        ret = zint_handlers[i]->resolve_path(
            path, newhandle, newpath, newpath_maxlen);
        if(ret == ZFS_OK)
        {
            return zint_handlers[i];
        }
        else if(ret != ZFSERR_NOENT)
        {
            return NULL;
        }
    }
    return NULL;
}

int zint_locate_handler(const zoidfs_handle_t * handle,
                        const char * path,
                        zint_handler_t ** handler,
                        char * newpath,
                        int newpath_maxlen,
                        zoidfs_handle_t * newhandle)
{
    if(handle)
    {
        *handler = zint_get_handler(ZOIDFS_HANDLE_TYPE(handle));
        if(!*handler)
        {
            return -1;
        }

        if(path)
        {
            strncpy(newpath, path, newpath_maxlen);
        }

        if(newhandle)
        {
            *newhandle = *handle;
        }
    }
    else if(path)
    {
        *handler = zint_get_handler_from_path(
            path, newpath, newpath_maxlen, newhandle);
        if(!*handler)
        {
            return -1;
        }
    }
    else
    {
        return -1;
    }

    return 0;
}

/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 * End:
 *
 * vim: ts=8 sts=4 sw=4 expandtab
 */
